import datetime
import logging
import socket
from pathlib import Path
from typing import Iterable, Optional

from cg.apps.demultiplex.demultiplex_api import DemultiplexingAPI
from cg.models.demultiplex.flowcell import Flowcell
from pydantic import BaseModel
from typing_extensions import Literal

LOG = logging.getLogger(__name__)


class LogfileParameters(BaseModel):
    id_string: str  # This indicate software and version
    # This is the binary that was executed (atm only bcl2fastq)
    program: Literal["bcl2fastq"] = "bcl2fastq"
    command_line: str
    time: datetime.datetime


class DemuxResults:
    """Class to gather information from a demultiplex result"""

    def __init__(self, demux_dir: Path, flowcell: Flowcell):
        LOG.info("Instantiating DemuxResults with path %s", demux_dir)
        self.demux_dir: Path = demux_dir
        self.flowcell: Flowcell = flowcell

    @property
    def run_name(self) -> str:
        return self.demux_dir.name

    @property
    def run_date(self) -> datetime.date:
        raw_time: str = self.run_name.split("_")[0]
        time_object: datetime.datetime = datetime.datetime.strptime(raw_time, "%y%m%d")
        return time_object.date()

    @property
    def machine_name(self) -> str:
        return self.run_name.split("_")[1]

    @property
    def demux_host(self) -> str:
        return socket.gethostname()

    @property
    def conversion_stats_path(self) -> Path:
        return self.results_dir / "Stats" / "ConversionStats.xml"

    @property
    def demux_stats_path(self) -> Path:
        return self.results_dir / "Stats" / "DemultiplexingStats.xml"

    @property
    def log_path(self) -> Path:
        return DemultiplexingAPI.get_logfile(flowcell=self.flowcell)

    @property
    def results_dir(self) -> Path:
        return self.demux_dir / "Unaligned"

    @property
    def sample_sheet_path(self) -> Path:
        return self.flowcell.sample_sheet_path

    @property
    def projects(self) -> Iterable[str]:
        """Return the project names created after demultiplexing

        How do we handle the Project_indexcheck directory?
        """
        for sub_dir in self.results_dir.iterdir():
            if not sub_dir.is_dir():
                continue
            if sub_dir.name.startswith("Project_"):
                project_name: str = sub_dir.name.split("_")[1]
                LOG.debug("Found project %s", project_name)
                yield project_name

    @property
    def raw_index_dir(self) -> Path:
        """Return the path to a index dir that is not given the 'Project_'-prefix"""
        return self.results_dir / "indexcheck"

    @property
    def raw_projects(self) -> Iterable[Path]:
        """Return the raw project names created after demultiplexing

        These are either 'indexcheck' or a number
        """
        for sub_dir in self.results_dir.iterdir():
            if not sub_dir.is_dir():
                LOG.debug("Skipping %s since it is not a directory", sub_dir)
                continue
            dir_name: str = sub_dir.name
            if dir_name in ["Stats", "Reports"]:
                LOG.debug("Skipping %s dir %s", dir_name, sub_dir)
                continue
            if dir_name.startswith("Project_"):
                LOG.debug("Skipping already renamed project dir %s", sub_dir)
                continue
            if "index" in dir_name:
                LOG.debug("Skipping 'indexcheck' dir %s", sub_dir)
                continue
                # We now know that the rest of the directories are project directories
            yield sub_dir

    def files_renamed(self) -> bool:
        """Assert if the files have been renamed

        Check if there are any raw projects, in that case it will need post processing
        """
        return bool(len(list(self.raw_projects)))

    def get_logfile_parameters(self) -> LogfileParameters:
        log_path: Path = self.log_path
        LOG.info("Parse log file %s", log_path)
        if not log_path.exists():
            LOG.warning("Could not find log file %s", log_path)
            raise FileNotFoundError
        program: str = ""
        time: Optional[datetime.datetime] = None
        command_line: Optional[str] = None
        id_string: Optional[str] = None
        with open(log_path, "r") as logfile:
            for line in logfile.readlines():
                # Fetch the line where the call that was made is
                if "bcl2fastq" in line and "singularity" in line:
                    line = line.strip()
                    split_line = line.split(" ")
                    command_line: str = " ".join(split_line[1:])
                    # Time is in format 2021-04-03-11:51:07, YYYY-MM-DD-HH-MM-SS
                    raw_time: str = split_line[0].strip("[]")  # remove the brackets around the date
                    try:
                        time: datetime.datetime = datetime.datetime.strptime(
                            raw_time, "%Y-%m-%dT%H:%M:%S"
                        )
                    except ValueError:
                        time: datetime.datetime = datetime.datetime.strptime(
                            raw_time, "%Y%m%d%H%M%S"
                        )
                    program = split_line[6]  # get the executed program

                if "bcl2fastq v" in line:
                    id_string = line.strip()
                    break

        return LogfileParameters(
            id_string=id_string, program=program, command_line=command_line, time=time
        )

    def __str__(self):
        return f"DemuxResults(demux_dir={self.demux_dir},flowcell=Flowcell(flowcell_path={self.flowcell.path})"



"""Hold sbatch templates in global variables"""

FASTQ_TO_SPRING_ERROR = """
if [[ -e {spring_path} ]]
then
    rm {spring_path}
fi

if [[ -e {pending_path} ]]
then
    rm {pending_path}
fi
"""

SPRING_TO_FASTQ_ERROR = """
if [[ -e {fastq_first} ]]
then
    rm {fastq_first}
fi

if [[ -e {fastq_second} ]]
then
    rm {fastq_second}
fi

if [[ -e {pending_path} ]]
then
    rm {pending_path}
fi
"""

FASTQ_TO_SPRING_COMMANDS = """
source activate {conda_env}

mkdir -p {tmp_dir}
crunchy -t 12 --tmp-dir {tmp_dir} compress fastq -f {fastq_first} -s {fastq_second} \
-o {spring_path} --check-integrity --metadata-file
rm {pending_path}
rm -r {tmp_dir}
"""

SPRING_TO_FASTQ_COMMANDS = """
source activate {conda_env}

mkdir -p {tmp_dir}
crunchy -t 12 --tmp-dir {tmp_dir} decompress spring {spring_path} -f {fastq_first} -s \
{fastq_second} --first-checksum {checksum_first} --second-checksum {checksum_second}
rm {pending_path}
rm -r {tmp_dir}
"""



from pathlib import Path
from typing import List

import pytest
from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.meta.upload.gisaid import GisaidAPI
from cg.meta.upload.gisaid.models import GisaidSample
from cg.models.cg_config import CGConfig, GisaidConfig
from cg.store import Store
from tests.store_helpers import StoreHelpers


@pytest.fixture(name="config")
def fixture_config(root_path) -> dict:
    """Return a dictionary with housekeeper api configs for testing"""
    return {"gisaid": {"binary_path": "/path/to/gisaid", "submitter": "some submitter"}}


@pytest.fixture(name="config_object_no_gisaid_samples")
def fixture_config_object_no_gisaid_samples(
    config: dict, cg_config_object: CGConfig, base_store: Store
):
    cg_config_object.status_db_ = base_store
    cg_config_object.gisaid = GisaidConfig(**config["gisaid"])
    return cg_config_object


@pytest.fixture(name="config_object")
def fixture_config_object(config: dict, cg_config_object: CGConfig, analysis_store: Store):
    cg_config_object.status_db_ = analysis_store
    cg_config_object.gisaid = GisaidConfig(**config["gisaid"])
    return cg_config_object


@pytest.fixture(scope="function")
def gisaid_api(config_object):
    """
    gisaid API fixture
    """
    return GisaidAPI(config=config_object)


@pytest.fixture(scope="function")
def gisaid_api_no_samples(config_object_no_gisaid_samples):
    """
    gisaid API fixture
    """
    return GisaidAPI(config=config_object_no_gisaid_samples)


@pytest.fixture
def valid_concat_fasta_file() -> str:
    """Get file path to valid fasta"""

    file_path = "tests/meta/upload/gisaid/fixtures/concat.fasta"
    file = Path(file_path)
    return file.read_text()


@pytest.fixture
def valid_fasta_file() -> str:
    """Get file path to valid fasta"""

    file_path = "tests/meta/upload/gisaid/fixtures/valid.fasta"
    file = Path(file_path)
    return str(file.absolute())


@pytest.fixture
def invalid_fasta_file() -> str:
    """Get file path to invalid fasta"""

    file_path = "tests/meta/upload/gisaid/fixtures/invalid.fasta"
    file = Path(file_path)
    return str(file.absolute())


@pytest.fixture
def gisaid_case_id():
    return "gisaid_case"


@pytest.fixture(name="gisaid_file_name")
def fixture_gisaid_file_name() -> Path:
    file_path = Path("test.fasta")
    yield file_path
    try:
        file_path.unlink()
    except FileNotFoundError:
        pass


@pytest.fixture
def four_gisaid_samples(gisaid_case_id: str, gisaid_file_name: Path) -> List[GisaidSample]:

    return [
        GisaidSample(
            family_id=gisaid_case_id,
            cg_lims_id=f"s{i}",
            covv_subm_sample_id="test_name",
            submitter="test_submitter",
            fn=gisaid_file_name.name,
            covv_collection_date="2020-11-22",
            lab="Stockholm",
        )
        for i in range(1, 5)
    ]


@pytest.fixture
def dummy_gisaid_sample(gisaid_file_name: Path) -> GisaidSample:

    return GisaidSample(
        family_id="dummy_id",
        cg_lims_id=f"dummy_sample",
        covv_subm_sample_id="test_name",
        submitter="test_submitter",
        fn=gisaid_file_name.name,
        covv_collection_date="2020-11-22",
        lab="Stockholm",
    )


@pytest.fixture
def bundle_with_four_samples(gisaid_case_id, timestamp):
    files = []
    for i in range(1, 5):
        file = Path(f"tests/meta/upload/gisaid/fixtures/s{i}_valid.fasta")
        files.append(
            {"path": str(file.absolute()), "archive": False, "tags": ["consensus", f"s{i}"]}
        )

    return {
        "name": gisaid_case_id,
        "created": timestamp,
        "expires": timestamp,
        "files": files,
    }


@pytest.fixture
def four_samples_csv() -> str:
    file_path = "tests/meta/upload/gisaid/fixtures/four_samples.csv"
    file = Path(file_path)
    return file.read_text()



import datetime


def is_similar_dicts(dict1, dict2):
    _is_similar = True

    for key in dict1.keys():
        _is_similar = _is_similar and is_similar_values(dict1.get(key), dict2.get(key))

    return _is_similar


def is_similar_lists(list1, list2):
    _is_similar = True

    if not list1:
        return list1 == list2

    if isinstance(list2, list):
        for value1, value2 in zip(list1, list2):
            _is_similar = _is_similar and is_similar_values(value1, value2)
    else:
        for value1 in list1:
            _is_similar = _is_similar and value1 in list2

    return _is_similar


def is_similar_values(value1, value2):

    if isinstance(value1, dict):
        return is_similar_dicts(value1, value2)

    if isinstance(value1, list):
        return is_similar_lists(value1, value2)

    if str(value1) == str(value2):
        return True

    if isinstance(value1, datetime.datetime):
        return str(value1.date()) == value2

    if is_float(value1):
        return is_float(value2) and round(float(value1), 1) == round(float(value2), 1)

    return False


def is_float(value):
    try:
        float(value)
        return True
    except (ValueError, TypeError):
        return False


def dict_values_exists_in(a_dict: dict, a_target: str):

    all_exists = True

    for value in a_dict.values():
        all_exists = all_exists and value_exists_in(value, a_target)
    return all_exists


def value_exists_in(value, a_target: str):

    if isinstance(value, str):
        return value in a_target
    if isinstance(value, float):
        return str(round(value, 2)) in a_target or str(round(value, 1)) in a_target
    if isinstance(value, dict):
        return dict_values_exists_in(value, a_target)
    if isinstance(value, list):
        return list_values_exists_in(value, a_target)
    if isinstance(value, datetime.datetime):
        return str(value.date()) in a_target

    if str(value) in a_target:
        return True

    if isinstance(value, bool):
        return True

    return False


def list_values_exists_in(a_list: list, a_target: str):

    all_exists = True

    for value in a_list:
        all_exists = all_exists and value_exists_in(value, a_target)
    return all_exists



import datetime as dt
import logging
import os
from pathlib import Path
from subprocess import CalledProcessError
from typing import List, Optional, Tuple

from cg.apps.environ import environ_email
from cg.constants import CASE_ACTIONS, Pipeline
from cg.exc import BundleAlreadyAddedError, CgDataError, CgError, DecompressionNeededError
from cg.meta.meta import MetaAPI
from cg.meta.workflow.fastq import FastqHandler
from cg.models.cg_config import CGConfig
from cg.store import Store, models
from housekeeper.store.models import Bundle, Version

LOG = logging.getLogger(__name__)


class AnalysisAPI(MetaAPI):
    """
    Parent class containing all methods that are either shared or overridden by other workflow APIs.
    """

    def __init__(self, pipeline: Pipeline, config: CGConfig):
        super().__init__(config=config)
        self.pipeline = pipeline
        self._process = None

    @property
    def threshold_reads(self):
        """Defines whether the threshold for adequate read count should be passed for all samples
        when determining if the analysis for a case should be automatically started"""
        return False

    @property
    def process(self):
        raise NotImplementedError

    @property
    def fastq_handler(self):
        return FastqHandler

    def verify_deliverables_file_exists(self, case_id: str) -> None:
        if not Path(self.get_deliverables_file_path(case_id=case_id)).exists():
            raise CgError(f"No deliverables file found for case {case_id}")

    def verify_case_config_file_exists(self, case_id: str):
        if not Path(self.get_case_config_path(case_id=case_id)).exists():
            raise CgError(f"No config file found for case {case_id}")

    def verify_analysis_finish_file_exists(self, case_id: str):
        if not Path(self.get_analysis_finish_path(case_id=case_id)).exists():
            raise CgError(f"No analysis_finish file found for case {case_id}")

    def verify_case_id_in_statusdb(self, case_id: str) -> None:
        """Passes silently if case exists in StatusDB, raises error if case is missing"""

        case_obj: models.Family = self.status_db.family(case_id)
        if not case_obj:
            LOG.error("Case %s could not be found in StatusDB!", case_id)
            raise CgError
        elif not case_obj.links:
            LOG.error("Case %s has no samples in in StatusDB!", case_id)
            raise CgError
        LOG.info("Case %s exists in status db", case_id)

    def check_analysis_ongoing(self, case_id: str) -> None:
        if self.trailblazer_api.is_latest_analysis_ongoing(case_id=case_id):
            LOG.warning(f"{case_id} : analysis is still ongoing - skipping")
            raise CgError(f"Analysis still ongoing in Trailblazer for case {case_id}")

    def verify_case_path_exists(self, case_id: str) -> None:
        if not self.get_case_path(case_id=case_id).exists():
            LOG.error("Working directory path for %s does not exist", case_id)
            raise CgError()

    def all_flowcells_on_disk(self, case_id: str) -> bool:
        """Check if flowcells are on disk for sample before starting the analysis.
        Flowcells not on disk will be requested
        """
        flowcells = self.status_db.flowcells(family=self.status_db.family(case_id))
        statuses = []
        for flowcell_obj in flowcells:
            LOG.info(f"{flowcell_obj.name}: checking if flowcell is on disk")
            LOG.info(f"{flowcell_obj.name}: status is {flowcell_obj.status}")
            statuses.append(flowcell_obj.status or "ondisk")
            if flowcell_obj.status == "removed":
                LOG.info(f"{flowcell_obj.name}: flowcell not on disk, requesting")
                flowcell_obj.status = "requested"
            elif flowcell_obj.status != "ondisk":
                LOG.warning(f"{flowcell_obj.name}: {flowcell_obj.status}")
        self.status_db.commit()
        return all(status == "ondisk" for status in statuses)

    def get_priority_for_case(self, case_id: str) -> str:
        """Fetch priority for case id"""
        case_obj: models.Family = self.status_db.family(case_id)
        if not case_obj.priority or case_obj.priority == 0:
            return "low"
        if case_obj.priority > 1:
            return "high"
        return "normal"

    def get_case_path(self, case_id: str) -> Path:
        """Path to case working directory"""
        raise NotImplementedError

    def get_case_config_path(self, case_id) -> Path:
        """Path to case config file"""
        raise NotImplementedError

    def get_trailblazer_config_path(self, case_id: str) -> Path:
        """Path to Trailblazer job id file"""
        raise NotImplementedError

    def get_sample_name_from_lims_id(self, lims_id: str) -> str:
        """Retrieve sample name provided by customer for specific sample"""
        sample_obj: models.Sample = self.status_db.sample(lims_id)
        return sample_obj.name

    def link_fastq_files(self, case_id: str, dry_run: bool = False) -> None:
        """
        Links fastq files from Housekeeper to case working directory
        """
        raise NotImplementedError

    def get_bundle_deliverables_type(self, case_id: str) -> Optional[str]:
        return None

    @staticmethod
    def get_application_type(sample_obj: models.Sample) -> str:
        """
        Gets application type for sample. Only application types supported by trailblazer (or other)
        are valid outputs
        """
        analysis_type: str = sample_obj.application_version.application.prep_category
        if analysis_type and analysis_type.lower() in [
            "wgs",
            "wes",
            "tgs",
        ]:
            return analysis_type.lower()
        return "other"

    def upload_bundle_housekeeper(self, case_id: str) -> None:
        """Storing bundle data in Housekeeper for CASE_ID"""

        LOG.info(f"Storing bundle data in Housekeeper for {case_id}")
        bundle_result: Tuple[Bundle, Version] = self.housekeeper_api.add_bundle(
            bundle_data=self.get_hermes_transformed_deliverables(case_id)
        )
        if not bundle_result:
            raise BundleAlreadyAddedError("Bundle already added to Housekeeper!")
        bundle_object, bundle_version = bundle_result
        self.housekeeper_api.include(bundle_version)
        self.housekeeper_api.add_commit(bundle_object, bundle_version)
        LOG.info(
            f"Analysis successfully stored in Housekeeper: {case_id} : {bundle_version.created_at}"
        )

    def upload_bundle_statusdb(self, case_id: str) -> None:
        """Storing analysis bundle in StatusDB for CASE_ID"""

        LOG.info(f"Storing analysis in StatusDB for {case_id}")
        case_obj: models.Family = self.status_db.family(case_id)
        analysis_start: dt.datetime = self.get_bundle_created_date(case_id=case_id)
        pipeline_version: str = self.get_pipeline_version(case_id=case_id)
        new_analysis: models.Family = self.status_db.add_analysis(
            pipeline=self.pipeline,
            version=pipeline_version,
            started_at=analysis_start,
            completed_at=dt.datetime.now(),
            primary=(len(case_obj.analyses) == 0),
        )
        new_analysis.family = case_obj
        self.status_db.add_commit(new_analysis)
        LOG.info(f"Analysis successfully stored in StatusDB: {case_id} : {analysis_start}")

    def get_deliverables_file_path(self, case_id: str) -> Path:
        raise NotImplementedError

    def get_analysis_finish_path(self, case_id: str) -> Path:
        raise NotImplementedError

    def add_pending_trailblazer_analysis(self, case_id: str) -> None:
        self.check_analysis_ongoing(case_id=case_id)
        self.trailblazer_api.mark_analyses_deleted(case_id=case_id)
        self.trailblazer_api.add_pending_analysis(
            case_id=case_id,
            email=environ_email(),
            analysis_type=self.get_application_type(self.status_db.family(case_id).links[0].sample),
            out_dir=self.get_trailblazer_config_path(case_id=case_id).parent.as_posix(),
            config_path=self.get_trailblazer_config_path(case_id=case_id).as_posix(),
            priority=self.get_priority_for_case(case_id=case_id),
            data_analysis=str(self.pipeline),
        )

    def get_hermes_transformed_deliverables(self, case_id: str) -> dict:
        return self.hermes_api.create_housekeeper_bundle(
            bundle_name=case_id,
            deliverables=self.get_deliverables_file_path(case_id=case_id),
            pipeline=str(self.pipeline),
            analysis_type=self.get_bundle_deliverables_type(case_id),
            created=self.get_bundle_created_date(case_id),
        ).dict()

    def get_bundle_created_date(self, case_id: str) -> dt.datetime:
        return self.get_date_from_file_path(self.get_deliverables_file_path(case_id=case_id))

    def get_pipeline_version(self, case_id: str) -> str:
        """
        Calls the pipeline to get the pipeline version number. If fails, returns a placeholder value instead.
        """
        try:
            self.process.run_command(["--version"])
            return list(self.process.stdout_lines())[0].split()[-1]
        except (Exception, CalledProcessError):
            LOG.warning("Could not retrieve %s workflow version!", self.pipeline)
            return "0.0.0"

    def set_statusdb_action(self, case_id: str, action: Optional[str]) -> None:
        """
        Set one of the allowed actions on a case in StatusDB.
        """
        if action in [None, *CASE_ACTIONS]:
            case_obj: models.Family = self.status_db.family(case_id)
            case_obj.action = action
            self.status_db.commit()
            LOG.info("Action %s set for case %s", action, case_id)
            return
        LOG.warning(
            f"Action '{action}' not permitted by StatusDB and will not be set for case {case_id}"
        )

    def get_analyses_to_clean(self, before: dt.datetime) -> List[models.Analysis]:
        analyses_to_clean = self.status_db.analyses_to_clean(pipeline=self.pipeline, before=before)
        return analyses_to_clean.all()

    def get_cases_to_analyze(self) -> List[models.Family]:
        return self.status_db.cases_to_analyze(
            pipeline=self.pipeline, threshold=self.threshold_reads
        )

    def get_running_cases(self) -> List[models.Family]:
        return self.status_db.get_running_cases_for_pipeline(pipeline=self.pipeline)

    def get_cases_to_store(self) -> List[models.Family]:
        raise NotImplementedError

    def get_sample_fastq_destination_dir(self, case_obj: models.Family, sample_obj: models.Sample):
        raise NotImplementedError

    def gather_file_metadata_for_sample(self, sample_obj: models.Sample) -> List[dict]:
        return [
            self.fastq_handler.parse_file_data(file_obj.full_path)
            for file_obj in self.housekeeper_api.files(
                bundle=sample_obj.internal_id, tags=["fastq"]
            )
        ]

    def link_fastq_files_for_sample(
        self, case_obj: models.Family, sample_obj: models.Sample, concatenate: bool = False
    ) -> None:
        """
        Link FASTQ files for a sample to working directory.
        If pipeline input requires concatenated fastq, files can also be concatenated
        """
        linked_reads_paths = {1: [], 2: []}
        concatenated_paths = {1: "", 2: ""}
        files: List[dict] = self.gather_file_metadata_for_sample(sample_obj=sample_obj)
        sorted_files = sorted(files, key=lambda k: k["path"])
        fastq_dir = self.get_sample_fastq_destination_dir(case_obj=case_obj, sample_obj=sample_obj)
        fastq_dir.mkdir(parents=True, exist_ok=True)

        for fastq_data in sorted_files:
            fastq_path = Path(fastq_data["path"])
            fastq_name = self.fastq_handler.create_fastq_name(
                lane=fastq_data["lane"],
                flowcell=fastq_data["flowcell"],
                sample=sample_obj.internal_id,
                read=fastq_data["read"],
                undetermined=fastq_data["undetermined"],
                meta=self.get_additional_naming_metadata(sample_obj),
            )
            destination_path: Path = fastq_dir / fastq_name
            linked_reads_paths[fastq_data["read"]].append(destination_path)
            concatenated_paths[
                fastq_data["read"]
            ] = f"{fastq_dir}/{self.fastq_handler.get_concatenated_name(fastq_name)}"

            if not destination_path.exists():
                LOG.info(f"Linking: {fastq_path} -> {destination_path}")
                destination_path.symlink_to(fastq_path)
            else:
                LOG.warning(f"Destination path already exists: {destination_path}")

        if not concatenate:
            return

        LOG.info("Concatenation in progress for sample %s.", sample_obj.internal_id)
        for read, value in linked_reads_paths.items():
            self.fastq_handler.concatenate(linked_reads_paths[read], concatenated_paths[read])
            self.fastq_handler.remove_files(value)

    def get_target_bed_from_lims(self, case_id: str) -> Optional[str]:
        """Get target bed filename from lims"""
        case_obj: models.Family = self.status_db.family(case_id)
        sample_obj: models.Sample = case_obj.links[0].sample
        if sample_obj.from_sample:
            sample_obj = self.status_db.sample(internal_id=sample_obj.internal_id)
        target_bed_shortname: str = self.lims_api.capture_kit(sample_obj.internal_id)
        if not target_bed_shortname:
            return target_bed_shortname
        bed_version_obj: Optional[models.BedVersion] = self.status_db.bed_version(
            target_bed_shortname
        )
        if not bed_version_obj:
            raise CgDataError("Bed-version %s does not exist" % target_bed_shortname)
        return bed_version_obj.filename

    def resolve_decompression(self, case_id: str, dry_run: bool) -> None:
        """
        Handles decompression automatically for case.
        Raises error to interrupt the workflow execution if cannot be resolved immediately
        """
        decompression_needed: bool = self.prepare_fastq_api.is_spring_decompression_needed(case_id)

        if decompression_needed:
            LOG.info(
                "The analysis for %s could not start, decompression is needed",
                case_id,
            )
            decompression_possible: bool = (
                self.prepare_fastq_api.can_at_least_one_sample_be_decompressed(case_id)
            )
            if decompression_possible:
                possible_to_start_decompression: bool = (
                    self.prepare_fastq_api.can_at_least_one_decompression_job_start(
                        case_id, dry_run
                    )
                )
                if possible_to_start_decompression:
                    if not dry_run:
                        self.set_statusdb_action(case_id=case_id, action="analyze")
                    LOG.info(
                        "Decompression started for %s",
                        case_id,
                    )
                else:
                    LOG.warning(
                        "Decompression failed to start for %s",
                        case_id,
                    )
            else:
                is_decompression_running = self.prepare_fastq_api.is_spring_decompression_running(
                    case_id
                )
                if is_decompression_running:
                    LOG.info(
                        "Decompression is running for %s, analysis will be started when decompression is done",
                        case_id,
                    )
                    self.set_statusdb_action(case_id=case_id, action="analyze")
                else:
                    LOG.warning(
                        "Decompression can not be started for %s",
                        case_id,
                    )
            raise DecompressionNeededError("Workflow interrupted: decompression is not finished")

        if self.prepare_fastq_api.is_spring_decompression_running(case_id):
            self.set_statusdb_action(case_id=case_id, action="analyze")
            raise DecompressionNeededError("Workflow interrupted: decompression is running")

        LOG.info("Linking fastq files in housekeeper for case %s", case_id)
        self.prepare_fastq_api.check_fastq_links(case_id)

        if self.prepare_fastq_api.is_spring_decompression_needed(case_id):
            raise DecompressionNeededError("Workflow interrupted: decompression is not finished")

        LOG.info("Decompression for case %s not needed", case_id)

    @staticmethod
    def get_date_from_file_path(file_path: Path) -> dt.datetime.date:
        """
        Get date from deliverables path using date created metadata.
        """
        return dt.datetime.fromtimestamp(int(os.path.getctime(file_path)))

    def get_additional_naming_metadata(self, sample_obj: models.Sample) -> Optional[str]:
        return None



"""This api should handle everything around demultiplexing"""
import logging
from pathlib import Path
from typing import Optional

from cg.apps.demultiplex.sbatch import DEMULTIPLEX_COMMAND, DEMULTIPLEX_ERROR
from cg.apps.slurm.slurm_api import SlurmAPI
from cg.models.demultiplex.flowcell import Flowcell
from cg.models.demultiplex.sbatch import SbatchCommand, SbatchError
from cg.models.slurm.sbatch import Sbatch
from typing_extensions import Literal

LOG = logging.getLogger(__name__)


class DemultiplexingAPI:
    """Demultiplexing API should deal with anything related to demultiplexing

    This includes starting demultiplexing, creating sample sheets, creating base masks
    """

    def __init__(self, config: dict, out_dir: Optional[Path] = None):
        self.slurm_api = SlurmAPI()
        self.slurm_account: str = config["demultiplex"]["slurm"]["account"]
        self.mail: str = config["demultiplex"]["slurm"]["mail_user"]
        self.run_dir: Path = Path(config["demultiplex"]["run_dir"])
        self.out_dir: Path = out_dir or Path(config["demultiplex"]["out_dir"])
        self.environment: str = config.get("environment", "stage")
        LOG.info("Set environment to %s", self.environment)
        self.dry_run: bool = False

    def set_dry_run(self, dry_run: bool) -> None:
        LOG.debug("Set dry run to %s", dry_run)
        self.dry_run = dry_run
        self.slurm_api.set_dry_run(dry_run=dry_run)

    @staticmethod
    def get_sbatch_error(
        flowcell: Flowcell,
        email: str,
        demux_dir: Path,
    ) -> str:
        """Create the sbatch error string"""
        LOG.info("Creating the sbatch error string")
        error_parameters: SbatchError = SbatchError(
            flowcell_name=flowcell.flowcell_id,
            email=email,
            logfile=DemultiplexingAPI.get_logfile(flowcell=flowcell).as_posix(),
            demux_dir=demux_dir.as_posix(),
            demux_started=flowcell.demultiplexing_started_path.as_posix(),
        )
        return DEMULTIPLEX_ERROR.format(**error_parameters.dict())

    @staticmethod
    def get_sbatch_command(
        run_dir: Path,
        unaligned_dir: Path,
        sample_sheet: Path,
        demux_completed: Path,
        environment: Literal["production", "stage"] = "stage",
    ) -> str:
        LOG.info("Creating the sbatch command string")
        command_parameters: SbatchCommand = SbatchCommand(
            run_dir=run_dir.as_posix(),
            demux_dir=unaligned_dir.parent.as_posix(),
            unaligned_dir=unaligned_dir.as_posix(),
            sample_sheet=sample_sheet.as_posix(),
            demux_completed_file=demux_completed.as_posix(),
            environment=environment,
        )
        return DEMULTIPLEX_COMMAND.format(**command_parameters.dict())

    @staticmethod
    def demultiplex_sbatch_path(flowcell: Flowcell) -> Path:
        """Get the path to where sbatch file should be kept"""
        return flowcell.path / "demux-novaseq.sh"

    @staticmethod
    def get_run_name(flowcell: Flowcell) -> str:
        """Create the run name for the sbatch job"""
        return f"{flowcell.flowcell_id}_demultiplex"

    @staticmethod
    def get_logfile(flowcell: Flowcell) -> Path:
        """Create the path to the logfile"""
        return flowcell.path / f"{DemultiplexingAPI.get_run_name(flowcell)}.stderr"

    def flowcell_out_dir_path(self, flowcell: Flowcell) -> Path:
        """Create the path to where the demuliplexed result should be produced"""
        return self.out_dir / flowcell.path.name

    def unaligned_dir_path(self, flowcell: Flowcell) -> Path:
        """Create the path to where the demuliplexed result should be produced"""
        return self.flowcell_out_dir_path(flowcell) / "Unaligned"

    def demultiplexing_completed_path(self, flowcell: Flowcell) -> Path:
        """Create the path to where the demuliplexed result should be produced"""
        return self.flowcell_out_dir_path(flowcell) / "demuxcomplete.txt"

    def is_demultiplexing_completed(self, flowcell: Flowcell) -> bool:
        """Create the path to where the demuliplexed result should be produced"""
        LOG.info("Check if demultiplexing is ready for %s", flowcell.path)
        logfile: Path = self.get_logfile(flowcell)
        if not logfile.exists():
            LOG.warning("Could not find logfile!")
            return False
        return self.demultiplexing_completed_path(flowcell).exists()

    def is_demultiplexing_ongoing(self, flowcell: Flowcell) -> bool:
        """Check if demultiplexing is ongoing

        This is indicated by if the file demuxstarted.txt exists in the flowcell directory
        AND
        that the demultiplexing completed file does not exist
        """
        LOG.debug("Check if demultiplexing is ongoing for %s", flowcell.flowcell_id)
        if not flowcell.demultiplexing_started_path.exists():
            LOG.debug("Demultiplexing has not been started")
            return False
        LOG.debug("Demultiplexing has been started!")
        if self.is_demultiplexing_completed(flowcell):
            LOG.debug("Demultiplexing is already completed for flowcell %s", flowcell.flowcell_id)
            return False
        LOG.debug("Demultiplexing is not finished!")
        return True

    def is_demultiplexing_possible(self, flowcell: Flowcell) -> bool:
        """Check if it is possible to start demultiplexing

        This means that
            - flowcell should be ready for demultiplexing (all files in place)
            - sample sheet needs to exist
            - demultiplexing should not be running
        """
        LOG.info("Check if demultiplexing is possible for %s", flowcell.flowcell_id)
        demultiplexing_possible = True
        if not flowcell.is_flowcell_ready():
            demultiplexing_possible = False

        if not flowcell.sample_sheet_exists():
            LOG.warning("Could not find sample sheet for %s", flowcell.flowcell_id)
            demultiplexing_possible = False

        if flowcell.is_demultiplexing_started():
            LOG.warning("Demultiplexing has already been started")
            demultiplexing_possible = False

        if self.flowcell_out_dir_path(flowcell=flowcell).exists():
            LOG.warning("Flowcell out dir exists")
            demultiplexing_possible = False

        if self.is_demultiplexing_completed(flowcell):
            LOG.warning("Demultiplexing is already completed for flowcell %s", flowcell.flowcell_id)
            demultiplexing_possible = False
        return demultiplexing_possible

    def create_demultiplexing_started_file(self, demultiplexing_started_path: Path) -> None:
        LOG.info("Creating demultiplexing started file")
        if self.dry_run:
            return
        demultiplexing_started_path.touch(exist_ok=False)

    def start_demultiplexing(self, flowcell: Flowcell):
        """Start demultiplexing for a flowcell"""
        self.create_demultiplexing_started_file(flowcell.demultiplexing_started_path)
        demux_dir: Path = self.flowcell_out_dir_path(flowcell=flowcell)
        unaligned_dir: Path = self.unaligned_dir_path(flowcell=flowcell)
        LOG.info("Demultiplexing to %s", unaligned_dir)
        if not self.dry_run:
            LOG.info("Creating demux dir %s", unaligned_dir)
            unaligned_dir.mkdir(exist_ok=False, parents=True)

        log_path: Path = self.get_logfile(flowcell=flowcell)
        error_function: str = self.get_sbatch_error(
            flowcell=flowcell, email=self.mail, demux_dir=demux_dir
        )
        commands: str = self.get_sbatch_command(
            run_dir=flowcell.path,
            unaligned_dir=unaligned_dir,
            sample_sheet=flowcell.sample_sheet_path,
            demux_completed=self.demultiplexing_completed_path(flowcell=flowcell),
            environment=self.environment,
        )

        sbatch_content: str = self.slurm_api.generate_sbatch_content(
            sbatch_parameters=Sbatch(
                job_name=self.get_run_name(flowcell),
                account=self.slurm_account,
                number_tasks=18,
                memory=50,
                log_dir=log_path.parent.as_posix(),
                email=self.mail,
                hours=36,
                commands=commands,
                error=error_function,
            )
        )
        sbatch_path: Path = self.demultiplex_sbatch_path(flowcell=flowcell)
        sbatch_number: int = self.slurm_api.submit_sbatch(
            sbatch_content=sbatch_content, sbatch_path=sbatch_path
        )
        LOG.info("Demultiplexing running as job %s", sbatch_number)
        return sbatch_number



"""Fixtures for cli tests"""

import pytest
from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.models.cg_config import CGConfig
from cg.store import Store
from click.testing import CliRunner


@pytest.fixture(name="cli_runner")
def fixture_cli_runner() -> CliRunner:
    """Create a CliRunner"""
    return CliRunner()


@pytest.fixture(name="application_tag")
def fixture_application_tag() -> str:
    """Create a CliRunner"""
    return "dummy_tag"


@pytest.fixture(name="base_context")
def fixture_base_context(
    base_store: Store, housekeeper_api: HousekeeperAPI, cg_config_object: CGConfig
) -> CGConfig:
    """context to use in cli"""
    cg_config_object.status_db_ = base_store
    cg_config_object.housekeeper_api_ = housekeeper_api
    return cg_config_object


@pytest.fixture(name="disk_store")
def fixture_disk_store(base_context: CGConfig) -> Store:
    """context to use in cli"""
    return base_context.status_db



"""Tests for rsync API"""
import glob
import logging

from cg.cli.deliver.base import rsync
from cg.constants import EXIT_SUCCESS
from cg.models.cg_config import CGConfig
from cgmodels.cg.constants import Pipeline


def test_rsync_help(cli_runner):
    """Test to run the rsync function"""
    # GIVEN a cli runner

    # WHEN running cg deliver rsync
    result = cli_runner.invoke(rsync, ["--dry-run", "--help"])

    # THEN assert the command exists without problems
    assert result.exit_code == EXIT_SUCCESS
    # THEN assert the information is printed
    assert "The folder generated using the" in result.output


def test_run_rsync_command(cg_context: CGConfig, cli_runner, helpers, caplog, mocker):
    """Test generating the rsync command for covid ticket"""
    caplog.set_level(logging.INFO)

    # Given a valid Sars-cov-2 case
    case = helpers.add_case(
        store=cg_context.status_db,
        internal_id="angrybird",
        case_id=999999,
        data_analysis=Pipeline.SARS_COV_2,
    )
    sample = helpers.add_sample(store=cg_context.status_db, ticket=999999)
    helpers.add_relationship(store=cg_context.status_db, sample=sample, case=case)
    cg_context.status_db.add_commit()

    # GIVEN file exists
    mocker.patch.object(glob, "glob")
    glob.glob.return_value = [
        "/folder_structure/angrybird/yet_another_folder/filename_999999_data_999.csv"
    ]

    # WHEN running deliver rsync command
    result = cli_runner.invoke(rsync, ["--dry-run", "999999"], obj=cg_context)

    # THEN command executed successfully
    assert result.exit_code == EXIT_SUCCESS

    # THEN process generates command string for linking analysis files
    assert "cust000/inbox/999999/ server.name.se:/some/cust000/path/999999/" in caplog.text
    # THEN process generates command string for linking report file
    assert (
        "rsync -rvL /folder_structure/angrybird/yet_another_folder/filename_999999_data_999.csv "
        "server.name.se:/another/cust000/foldername/" in caplog.text
    )


def test_run_rsync_command_no_file(cg_context: CGConfig, cli_runner, helpers, caplog, mocker):
    """Test generating the rsync command for covid ticket"""
    caplog.set_level(logging.INFO)

    # Given a valid Sars-cov-2 case
    case = helpers.add_case(
        store=cg_context.status_db,
        internal_id="angrybird",
        case_id=999999,
        data_analysis=Pipeline.SARS_COV_2,
    )
    sample = helpers.add_sample(store=cg_context.status_db, ticket=999999)
    helpers.add_relationship(store=cg_context.status_db, sample=sample, case=case)
    cg_context.status_db.add_commit()

    # WHEN running deliver rsync command
    result = cli_runner.invoke(rsync, ["--dry-run", "999999"], obj=cg_context)

    # THEN command executed successfully
    assert result.exit_code == EXIT_SUCCESS

    # THEN process generates command string for linking analysis files
    assert "cust000/inbox/999999/ server.name.se:/some/cust000/path/999999/" in caplog.text
    # THEN process generates command string for linking report file
    caplog.set_level(logging.ERROR)
    assert "No report file could be found" in caplog.text


def test_run_rsync_command_no_case(cg_context: CGConfig, cli_runner, helpers, caplog):
    """Test generating the rsync command for ticket that doesnt exist"""
    caplog.set_level(logging.INFO)

    # Given an invalid ticket id where case was not created

    # WHEN running deliver rsync command
    result = cli_runner.invoke(rsync, ["--dry-run", "9898989898"], obj=cg_context)

    # THEN command failed successfully
    assert result.exit_code != EXIT_SUCCESS

    # THEN process generates error message that case cant be found
    assert "Could not find any cases for ticket_id" in caplog.text


def test_run_rsync_command_not_sarscov(cg_context: CGConfig, cli_runner, helpers, caplog):
    """Test generating the rsync command for non-covid ticket"""
    caplog.set_level(logging.INFO)

    # Given a valid Sars-cov-2 case
    case = helpers.add_case(
        store=cg_context.status_db,
        internal_id="angryfox",
        case_id=987654,
        data_analysis=Pipeline.BALSAMIC,
    )
    sample = helpers.add_sample(store=cg_context.status_db, ticket=987654)
    helpers.add_relationship(store=cg_context.status_db, sample=sample, case=case)
    cg_context.status_db.add_commit()

    # WHEN running deliver rsync command
    result = cli_runner.invoke(rsync, ["--dry-run", "987654"], obj=cg_context)

    # THEN command executed successfully
    assert result.exit_code == EXIT_SUCCESS

    # THEN process will not inform about linking report file
    assert "Delivering report for SARS-COV-2 analysis" not in caplog.text



import logging
from pathlib import Path

LOG = logging.getLogger(__name__)


def rename_index_dir(unaligned_dir: Path, dry_run: bool = False) -> None:
    """Rename the index directory by adding the prefix Project_"""
    for sub_dir in unaligned_dir.iterdir():
        if not sub_dir.is_dir():
            continue
        if sub_dir.name != "indexcheck":
            continue
        if sub_dir.name.startswith("Project_"):
            LOG.debug("%s is already renamed", sub_dir)
            continue
        index_directory: Path = unaligned_dir / "_".join(["Project", sub_dir.name])
        LOG.info("Move index directory %s", sub_dir)
        if not dry_run:
            LOG.info("Move directory to %s", index_directory)
            sub_dir.rename(index_directory)


def rename_project_directory(
    project_directory: Path, flowcell_id: str, dry_run: bool = False
) -> None:
    """Rename a project directory by adding the prefix Project_"""
    unaligned_directory: Path = project_directory.parent
    LOG.info("Rename all sample directories in %s", unaligned_directory)
    for sample_dir in project_directory.iterdir():
        if sample_dir.name.startswith("Sample"):
            LOG.debug("%s is already renamed", sample_dir)
            continue
        rename_sample_directory(
            sample_directory=sample_dir, flowcell_id=flowcell_id, dry_run=dry_run
        )
    new_name: str = "_".join(["Project", project_directory.name])
    new_project_path: Path = unaligned_directory / new_name
    LOG.info("Rename project dir %s", project_directory)
    if not dry_run:
        LOG.info("Rename project dir to %s", new_project_path)
        project_directory.rename(new_project_path)


def rename_sample_directory(
    sample_directory: Path, flowcell_id: str, dry_run: bool = False
) -> None:
    """Rename a sample dir and all fastq files in the sample dir

    Renaming of the sample dir means adding Sample_ as a prefix
    """
    project_directory: Path = sample_directory.parent
    LOG.info("Renaming all fastq files in %s", sample_directory)
    for fastq_file in sample_directory.iterdir():
        rename_fastq_file(fastq_file=fastq_file, flowcell_id=flowcell_id, dry_run=dry_run)
    new_name: str = "_".join(["Sample", sample_directory.name])
    new_sample_directory: Path = project_directory / new_name
    LOG.info("Renaming sample dir %s to %s", sample_directory, new_sample_directory)
    if not dry_run:
        sample_directory.rename(new_sample_directory)
        LOG.info("Renamed sample dir to %s", new_sample_directory)


def rename_fastq_file(fastq_file: Path, flowcell_id: str, dry_run: bool = False) -> None:
    """Rename a fastq file by appending the flowcell id as a prefix"""
    new_name: str = "_".join([flowcell_id, fastq_file.name])
    new_file: Path = Path(fastq_file.parent) / new_name
    LOG.info("Renaming fastq file %s to %s", fastq_file, new_file)
    if not dry_run:
        fastq_file.rename(new_file)
        LOG.info("Renamed fastq file to %s", new_file)



"""Tests for the deliver base command"""

from cg.cli.deliver.base import deliver as deliver_cmd
from cg.cli.deliver.base import rsync as rsync_cmd
from cg.constants import EXIT_SUCCESS
from cg.models.cg_config import CGConfig
from click.testing import CliRunner


def test_run_base_help():
    """Test to run the deliver base command with --help"""
    # GIVEN a cli runner
    runner = CliRunner()

    # WHEN running cg deliver help
    result = runner.invoke(deliver_cmd, ["--help"])

    # THEN assert the command exists without problems
    assert result.exit_code == EXIT_SUCCESS
    # THEN assert the information is printed
    assert "Deliver files with CG" in result.output


def test_run_deliver_analysis_help(base_context: CGConfig):
    """Test to run the deliver base command with --help"""
    # GIVEN a cli runner
    runner = CliRunner()
    # GIVEN a context with store and housekeeper information

    # WHEN running cg deliver help
    result = runner.invoke(deliver_cmd, ["analysis", "--help"], obj=base_context)

    # THEN assert the command exists without problems
    assert result.exit_code == EXIT_SUCCESS
    # THEN assert the information is printed
    assert "Deliver analysis files to customer inbox" in result.output



"""Set case attributes in the status database"""
import logging
from typing import Optional, Tuple

import click
from cg.constants import CASE_ACTIONS, PRIORITY_OPTIONS, DataDelivery, Pipeline
from cg.models.cg_config import CGConfig
from cg.store import Store, models
from cg.utils.click.EnumChoice import EnumChoice

LOG = logging.getLogger(__name__)


@click.command()
@click.option("-a", "--action", type=click.Choice(CASE_ACTIONS), help="update case action")
@click.option("-c", "--customer-id", type=click.STRING, help="update customer")
@click.option(
    "-d",
    "--data-analysis",
    "data_analysis",
    type=EnumChoice(Pipeline),
    help="Update case data analysis",
)
@click.option(
    "-dd",
    "--data-delivery",
    "data_delivery",
    type=EnumChoice(DataDelivery),
    help="Update case data delivery",
)
@click.option("-g", "--panel", "panels", multiple=True, help="update gene panels")
@click.option("-p", "--priority", type=click.Choice(PRIORITY_OPTIONS), help="update priority")
@click.argument("family_id")
@click.pass_obj
def family(
    context: CGConfig,
    action: Optional[str],
    data_analysis: Optional[Pipeline],
    data_delivery: Optional[DataDelivery],
    priority: Optional[str],
    panels: Optional[Tuple[str]],
    family_id: str,
    customer_id: Optional[str],
):
    """Update information about a case."""
    status_db: Store = context.status_db
    case_obj: models.Family = status_db.family(family_id)
    if case_obj is None:
        LOG.error("Can't find case %s,", family_id)
        raise click.Abort
    if not any([action, panels, priority, customer_id, data_analysis, data_delivery]):
        LOG.error("Nothing to change")
        raise click.Abort
    if action:
        LOG.info("Update action: %s -> %s", case_obj.action or "NA", action)
        case_obj.action = action
    if customer_id:
        customer_obj: models.Customer = status_db.customer(customer_id)
        if customer_obj is None:
            LOG.error("Unknown customer: %s", customer_id)
            raise click.Abort
        LOG.info(f"Update customer: {case_obj.customer.internal_id} -> {customer_id}")
        case_obj.customer = customer_obj
    if data_analysis:
        LOG.info(f"Update data_analysis: {case_obj.data_analysis or 'NA'} -> {data_analysis}")
        case_obj.data_analysis = data_analysis
    if data_delivery:
        LOG.info(f"Update data_delivery: {case_obj.data_delivery or 'NA'} -> {data_delivery}")
        case_obj.data_delivery = data_delivery
    if panels:
        for panel_id in panels:
            panel_obj: models.Panel = status_db.panel(panel_id)
            if panel_obj is None:
                LOG.error(f"unknown gene panel: {panel_id}")
                raise click.Abort
        LOG.info(f"Update panels: {', '.join(case_obj.panels)} -> {', '.join(panels)}")
        case_obj.panels = panels
    if priority:
        LOG.info(f"update priority: {case_obj.priority_human} -> {priority}")
        case_obj.priority_human = priority

    status_db.commit()



from pathlib import Path

from cg.apps.demultiplex.sample_sheet.novaseq_sample_sheet import SampleSheetCreator
from cgmodels.demultiplex.sample_sheet import SampleSheet, get_sample_sheet


def test_convert_to_sheet(novaseq_sample_sheet_object: SampleSheetCreator, project_dir: Path):
    # GIVEN a sample sheet object populated with samples
    assert novaseq_sample_sheet_object.lims_samples

    # WHEN converting to sample sheet
    sample_sheet: str = novaseq_sample_sheet_object.construct_sample_sheet()

    # THEN assert a correctly formatted sample sheet was created
    sample_sheet_object: SampleSheet = get_sample_sheet(sample_sheet=sample_sheet, sheet_type="S4")
    assert sample_sheet_object.samples



import logging
from pathlib import Path
from typing import List

import click
from cg.apps.demultiplex.demultiplex_api import DemultiplexingAPI
from cg.apps.demultiplex.sample_sheet.create import create_sample_sheet
from cg.apps.lims.samplesheet import LimsFlowcellSample, flowcell_samples
from cg.models.cg_config import CGConfig
from cg.models.demultiplex.flowcell import Flowcell
from cgmodels.demultiplex.sample_sheet import get_sample_sheet_from_file
from pydantic import ValidationError

LOG = logging.getLogger(__name__)


@click.group(name="samplesheet")
def sample_sheet_commands():
    """Command group for the sample sheet commands"""
    pass


@sample_sheet_commands.command(name="validate")
@click.argument("sheet", type=click.Path(exists=True, dir_okay=False))
def validate_sample_sheet(sheet: click.Path):
    """Command to validate a sample sheet"""
    LOG.info("Validating sample sheet %s", sheet)
    sheet: Path = Path(str(sheet))
    if sheet.suffix != ".csv":
        LOG.warning("File %s seems to be in wrong format", sheet)
        LOG.warning("Suffix %s is not '.csv'", sheet.suffix)
        raise click.Abort
    try:
        get_sample_sheet_from_file(infile=sheet, sheet_type="S4")
    except ValidationError as err:
        LOG.warning(err)
        raise click.Abort
    LOG.info("Sample sheet looks fine")


@sample_sheet_commands.command(name="create")
@click.argument("flowcell-name")
@click.option("--dry-run", is_flag=True)
@click.pass_obj
def create_sheet(context: CGConfig, flowcell_name: str, dry_run: bool):
    """Command to create a sample sheet

    Search the flowcell in the directory specified in config
    """
    LOG.info("Creating sample sheet for flowcell %s", flowcell_name)
    demultiplex_api: DemultiplexingAPI = context.demultiplex_api
    flowcell_path: Path = demultiplex_api.run_dir / flowcell_name
    if not flowcell_path.exists():
        LOG.warning("Could not find flowcell %s", flowcell_path)
        raise click.Abort
    flowcell_object = Flowcell(flowcell_path=flowcell_path)
    lims_samples: List[LimsFlowcellSample] = list(
        flowcell_samples(lims=context.lims_api, flowcell_id=flowcell_object.flowcell_id)
    )
    if not lims_samples:
        LOG.warning("Could not find any samples in lims for %s", flowcell_object.flowcell_id)
        raise click.Abort

    try:
        sample_sheet: str = create_sample_sheet(flowcell=flowcell_object, lims_samples=lims_samples)
    except (FileNotFoundError, FileExistsError):
        raise click.Abort

    if dry_run:
        click.echo(sample_sheet)
        return
    LOG.info("Writing sample sheet to %s", flowcell_object.sample_sheet_path.resolve())
    with open(flowcell_object.sample_sheet_path, "w") as outfile:
        outfile.write(sample_sheet)


@sample_sheet_commands.command(name="create-all")
@click.option("--dry-run", is_flag=True)
@click.pass_obj
def create_all_sheets(context: CGConfig, dry_run: bool):
    """Command to create sample sheets for all flowcells that lack a sample sheet

    Search flowcell directories for run parameters and create a sample sheets based on the information
    """
    demux_api: DemultiplexingAPI = context.demultiplex_api
    flowcells: Path = demux_api.run_dir
    for sub_dir in flowcells.iterdir():
        if not sub_dir.is_dir():
            continue
        LOG.info("Found directory %s", sub_dir)
        flowcell_object = Flowcell(flowcell_path=sub_dir)
        if flowcell_object.sample_sheet_exists():
            LOG.info("Sample sheet already exists")
            continue
        LOG.info("Creating sample sheet for flowcell %s", flowcell_object.flowcell_id)
        lims_samples: List[LimsFlowcellSample] = list(
            flowcell_samples(lims=context.lims_api, flowcell_id=flowcell_object.flowcell_id)
        )
        if not lims_samples:
            LOG.warning("Could not find any samples in lims for %s", flowcell_object.flowcell_id)
            continue

        try:
            sample_sheet: str = create_sample_sheet(
                flowcell=flowcell_object, lims_samples=lims_samples
            )
        except (FileNotFoundError, FileExistsError):
            continue

        if dry_run:
            click.echo(sample_sheet)
            return
        LOG.info("Writing sample sheet to %s", flowcell_object.sample_sheet_path.resolve())
        with open(flowcell_object.sample_sheet_path, "w") as outfile:
            outfile.write(sample_sheet)



"""Interactions with the gisaid cli upload"""

import logging
from pathlib import Path
from typing import List

from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.models.cg_config import CGConfig
from cg.store import models, Store
from cg.utils import Process
from .constants import HEADERS
from .models import GisaidSample, FastaFile
from housekeeper.store import models as hk_models
import csv

from cg.exc import HousekeeperVersionMissingError, FastaSequenceMissingError

LOG = logging.getLogger(__name__)


class GisaidAPI:
    """Interface with Gisaid cli uppload"""

    def __init__(self, config: CGConfig):
        self.housekeeper_api: HousekeeperAPI = config.housekeeper_api
        self.status_db: Store = config.status_db
        self.gisaid_submitter: str = config.gisaid.submitter
        self.gisaid_binary: str = config.gisaid.binary_path
        self.process = Process(binary=self.gisaid_binary)

    def get_fasta_sequence(self, fastq_path: str) -> str:
        """Get header from fasta file. Assuming un zipped file"""

        with open(fastq_path) as handle:
            fasta_lines = handle.readlines()
            try:
                return fasta_lines[1].rstrip("\n")
            except:
                raise FastaSequenceMissingError

    def get_gisaid_fasta_objects(self, gsaid_samples: List[GisaidSample]) -> List[FastaFile]:
        """Fetch a fasta files form house keeper for batch upload to gisaid"""

        fasta_objects = []
        for sample in gsaid_samples:
            hk_version: hk_models.Version = self.housekeeper_api.last_version(
                bundle=sample.family_id
            )
            if not hk_version:
                LOG.info("Family ID: %s not found in hose keeper", sample.family_id)
                raise HousekeeperVersionMissingError
            fasta_file: str = self.housekeeper_api.files(
                version=hk_version.id, tags=["consensus", sample.cg_lims_id]
            ).first()
            fasta_sequence: str = self.get_fasta_sequence(fastq_path=fasta_file.full_path)
            fasta_obj = FastaFile(header=sample.covv_virus_name, sequence=fasta_sequence)
            fasta_objects.append(fasta_obj)
        return fasta_objects

    def build_gisaid_fasta(self, gsaid_samples: List[GisaidSample], file_name: str) -> Path:
        """Concatenates a list of consensus fastq objects"""

        file: Path = Path(file_name)
        fasta_objects: List[FastaFile] = self.get_gisaid_fasta_objects(gsaid_samples=gsaid_samples)
        with open(file, "w") as write_file_obj:
            for fasta_object in fasta_objects:
                write_file_obj.write(f">{fasta_object.header}\n")
                write_file_obj.write(f"{fasta_object.sequence}\n")
        return file

    def get_sample_row(self, gisaid_sample: GisaidSample) -> List[str]:
        """Build row for a sample in the batch upload csv."""

        sample_dict = gisaid_sample.dict()
        return [sample_dict.get(header, "") for header in HEADERS]

    def get_sample_rows(self, gsaid_samples: List[GisaidSample]) -> List[List[str]]:
        """Build sample row list for gisad csv file"""

        sample_rows = []
        for sample in gsaid_samples:
            sample_row: List[str] = self.get_sample_row(gisaid_sample=sample)
            sample_rows.append(sample_row)
        return sample_rows

    def build_gisaid_csv(self, gsaid_samples: List[GisaidSample], file_name: str) -> Path:
        """Build batch upload csv."""

        file: Path = Path(file_name)
        with open(file_name, "w", newline="\n") as gisaid_csv:
            wr = csv.writer(gisaid_csv, delimiter=",")
            wr.writerow(HEADERS)
            sample_rows: List[List[str]] = self.get_sample_rows(gsaid_samples=gsaid_samples)
            wr.writerows(sample_rows)

        return file

    def get_gsaid_samples(self, family_id: str) -> List[GisaidSample]:
        """Get list of Gisaid sample objects."""

        samples: List[models.Sample] = self.status_db.get_sequenced_samples(family_id=family_id)
        gisaid_samples = []
        for sample in samples:
            gisaid_sample = GisaidSample(
                family_id=family_id,
                cg_lims_id=sample.internal_id,
                covv_subm_sample_id=sample.name,
                submitter=self.gisaid_submitter,
                fn=f"{family_id}.fasta",
                covv_collection_date="2020-11-22",  # sample.collection_date
                lab="Stockholm",  # sample.originating_lab,
            )
            gisaid_samples.append(gisaid_sample)
        return gisaid_samples

    def upload(self, csv_file: Path, fasta_file: Path) -> None:
        """Load batch data to GISAID using the gisiad cli."""

        load_call: list = [
            "CoV",
            "upload",
            "--csv",
            str(csv_file.absolute()),
            "--fasta",
            str(fasta_file.absolute()),
        ]
        self.process.run_command(parameters=load_call)

        if self.process.stderr:
            LOG.info(f"gisaid stderr:\n{self.process.stderr}")
        if self.process.stdout:
            LOG.info(f"gisaid stdout:\n{self.process.stdout}")

    def __str__(self):
        return f"GisaidAPI(dry_run: {self.dry_run})"



"""Fixtures for the upload scout api tests"""

import json
import logging
from datetime import datetime
from pathlib import Path
from typing import List

import pytest
from cg.constants import Pipeline
from cg.meta.upload.scout.balsamic_config_builder import BalsamicConfigBuilder
from cg.meta.upload.scout.mip_config_builder import MipConfigBuilder
from cg.meta.upload.scout.scoutapi import UploadScoutAPI
from cg.models.scout.scout_load_config import MipLoadConfig
from cg.store import Store, models
from housekeeper.store import models as hk_models

# Mocks
from tests.mocks.hk_mock import MockHousekeeperAPI
from tests.mocks.limsmock import MockLimsAPI
from tests.mocks.madeline import MockMadelineAPI
from tests.mocks.scout import MockScoutAPI
from tests.store_helpers import StoreHelpers

LOG = logging.getLogger(__name__)


class MockAnalysis:
    """Mock an analysis object"""

    @staticmethod
    def get_latest_metadata(family_id=None):
        """Mock get_latest_metadata"""
        # Returns: dict: parsed data
        # Define output dict
        out_data = {
            "analysis_sex": {"ADM1": "female", "ADM2": "female", "ADM3": "female"},
            "case": family_id or "yellowhog",
            "duplicates": {"ADM1": 13.525, "ADM2": 12.525, "ADM3": 14.525},
            "genome_build": "hg19",
            "rank_model_version": "1.18",
            "mapped_reads": {"ADM1": 98.8, "ADM2": 99.8, "ADM3": 97.8},
            "mip_version": "v4.0.20",
            "sample_ids": ["2018-20203", "2018-20204"],
            "sv_rank_model_version": "1.08",
        }
        return out_data

    @staticmethod
    def convert_panels(customer_id, panels):
        """Mock convert_panels"""
        _ = customer_id, panels
        return ""


@pytest.fixture(name="lims_family")
def fixture_lims_family() -> dict:
    """Returns a lims-like case of samples"""
    return json.load(open("tests/fixtures/report/lims_family.json"))


@pytest.fixture(name="lims_samples")
def fixture_lims_samples(lims_family: dict) -> List[dict]:
    """Returns the samples of a lims case"""
    return lims_family["samples"]


@pytest.fixture(scope="function", name="mip_analysis_hk_bundle_data")
def fixture_mip_analysis_hk_bundle_data(
    case_id: str, timestamp: datetime, mip_dna_analysis_dir: Path, sample_id: str
) -> dict:
    """Get some bundle data for housekeeper"""
    data = {
        "name": case_id,
        "created": timestamp,
        "expires": timestamp,
        "files": [
            {
                "path": str(mip_dna_analysis_dir / "snv.vcf"),
                "archive": False,
                "tags": ["vcf-snv-clinical"],
            },
            {
                "path": str(mip_dna_analysis_dir / "sv.vcf"),
                "archive": False,
                "tags": ["vcf-sv-clinical"],
            },
            {
                "path": str(mip_dna_analysis_dir / "snv_research.vcf"),
                "archive": False,
                "tags": ["vcf-snv-research"],
            },
            {
                "path": str(mip_dna_analysis_dir / "sv_research.vcf"),
                "archive": False,
                "tags": ["vcf-sv-research"],
            },
            {
                "path": str(mip_dna_analysis_dir / "str.vcf"),
                "archive": False,
                "tags": ["vcf-str"],
            },
            {
                "path": str(mip_dna_analysis_dir / "smn.vcf"),
                "archive": False,
                "tags": ["smn-calling"],
            },
            {
                "path": str(mip_dna_analysis_dir / "adm1.cram"),
                "archive": False,
                "tags": ["cram", sample_id],
            },
            {
                "path": str(mip_dna_analysis_dir / "report.pdf"),
                "archive": False,
                "tags": ["delivery-report"],
            },
            {
                "path": str(mip_dna_analysis_dir / "adm1.mt.bam"),
                "archive": False,
                "tags": ["bam-mt", sample_id],
            },
            {
                "path": str(mip_dna_analysis_dir / "vcf2cytosure.txt"),
                "archive": False,
                "tags": ["vcf2cytosure", sample_id],
            },
            {
                "path": str(mip_dna_analysis_dir / "multiqc.html"),
                "archive": False,
                "tags": ["multiqc-html", sample_id],
            },
        ],
    }
    return data


@pytest.fixture(scope="function", name="balsamic_analysis_hk_bundle_data")
def fixture_balsamic_analysis_hk_bundle_data(
    case_id: str, timestamp: datetime, balsamic_panel_analysis_dir: Path, sample_id: str
) -> dict:
    """Get some bundle data for housekeeper"""
    return {
        "name": case_id,
        "created": timestamp,
        "expires": timestamp,
        "files": [
            {
                "path": str(balsamic_panel_analysis_dir / "snv.vcf"),
                "archive": False,
                "tags": ["vcf-snv-clinical"],
            },
            {
                "path": str(balsamic_panel_analysis_dir / "sv.vcf"),
                "archive": False,
                "tags": ["vcf-sv-clinical"],
            },
            {
                "path": str(balsamic_panel_analysis_dir / "adm1.cram"),
                "archive": False,
                "tags": ["cram", sample_id],
            },
            {
                "path": str(balsamic_panel_analysis_dir / "coverage_qc_report.pdf"),
                "archive": False,
                "tags": ["delivery-report"],
            },
        ],
    }


@pytest.fixture(name="balsamic_analysis_hk_version")
def fixture_balsamic_analysis_hk_version(
    housekeeper_api: MockHousekeeperAPI, balsamic_analysis_hk_bundle_data: dict, helpers
) -> MockHousekeeperAPI:
    return helpers.ensure_hk_version(housekeeper_api, balsamic_analysis_hk_bundle_data)


@pytest.fixture(name="mip_analysis_hk_version")
def fixture_mip_analysis_hk_version(
    housekeeper_api: MockHousekeeperAPI, mip_analysis_hk_bundle_data: dict, helpers
) -> MockHousekeeperAPI:
    return helpers.ensure_hk_version(housekeeper_api, mip_analysis_hk_bundle_data)


@pytest.fixture(name="mip_analysis_hk_api")
def fixture_mip_analysis_hk_api(
    housekeeper_api: MockHousekeeperAPI, mip_analysis_hk_bundle_data: dict, helpers
) -> MockHousekeeperAPI:
    """Return a housekeeper api populated with some mip analysis files"""
    helpers.ensure_hk_version(housekeeper_api, mip_analysis_hk_bundle_data)
    return housekeeper_api


@pytest.fixture(name="balsamic_analysis_hk_api")
def fixture_balsamic_analysis_hk_api(
    housekeeper_api: MockHousekeeperAPI, balsamic_analysis_hk_bundle_data: dict, helpers
) -> MockHousekeeperAPI:
    """Return a housekeeper api populated with some mip analysis files"""
    helpers.ensure_hk_version(housekeeper_api, balsamic_analysis_hk_bundle_data)
    return housekeeper_api


@pytest.fixture(name="mip_file_handler")
def fixture_mip_file_handler(mip_analysis_hk_version: hk_models.Version) -> MipConfigBuilder:
    return MipConfigBuilder(hk_version_obj=mip_analysis_hk_version)


@pytest.fixture(name="mip_analysis_obj")
def fixture_mip_analysis_obj(
    analysis_store_trio: Store, case_id: str, timestamp: datetime, helpers: StoreHelpers
) -> models.Analysis:
    helpers.add_synopsis_to_case(store=analysis_store_trio, case_id=case_id)
    case_obj: models.Family = analysis_store_trio.family(case_id)
    analysis_obj: models.Analysis = helpers.add_analysis(
        store=analysis_store_trio,
        case=case_obj,
        started_at=timestamp,
        pipeline=Pipeline.MIP_DNA,
        completed_at=timestamp,
    )
    return analysis_obj


@pytest.fixture(name="balsamic_analysis_obj")
def fixture_balsamic_analysis_obj(analysis_obj: models.Analysis) -> models.Analysis:
    for link_object in analysis_obj.family.links:
        link_object.sample.application_version.application.prep_category = "wes"
    return analysis_obj


@pytest.fixture(name="mip_config_builder")
def fixture_mip_config_builder(
    mip_analysis_hk_version: hk_models.Version,
    mip_analysis_obj: models.Analysis,
    lims_api: MockLimsAPI,
    mip_analysis_api: MockAnalysis,
    madeline_api: MockMadelineAPI,
) -> MipConfigBuilder:
    return MipConfigBuilder(
        hk_version_obj=mip_analysis_hk_version,
        analysis_obj=mip_analysis_obj,
        lims_api=lims_api,
        mip_analysis_api=mip_analysis_api,
        madeline_api=madeline_api,
    )


@pytest.fixture(name="balsamic_config_builder")
def fixture_balsamic_config_builder(
    balsamic_analysis_hk_version: hk_models.Version,
    balsamic_analysis_obj: models.Analysis,
    lims_api: MockLimsAPI,
) -> BalsamicConfigBuilder:
    return BalsamicConfigBuilder(
        hk_version_obj=balsamic_analysis_hk_version,
        analysis_obj=balsamic_analysis_obj,
        lims_api=lims_api,
    )


@pytest.fixture(name="mip_load_config")
def fixture_mip_load_config(
    mip_dna_analysis_dir: Path, case_id: str, customer_id: str
) -> MipLoadConfig:
    """Return a valid mip load_config"""
    return MipLoadConfig(
        owner=customer_id,
        family=case_id,
        vcf_snv=str(mip_dna_analysis_dir / "snv.vcf"),
        track="rare",
    )


@pytest.fixture(name="lims_api")
def fixture_lims_api(lims_samples: List[dict]) -> MockLimsAPI:
    return MockLimsAPI(samples=lims_samples)


@pytest.fixture(name="mip_analysis_api")
def fixture_mip_analysis_api() -> MockAnalysis:
    return MockAnalysis()


@pytest.fixture(name="upload_scout_api")
def fixture_upload_scout_api(
    scout_api: MockScoutAPI,
    madeline_api: MockMadelineAPI,
    lims_samples: List[dict],
    housekeeper_api: MockHousekeeperAPI,
) -> UploadScoutAPI:
    """Fixture for upload_scout_api"""
    analysis_mock = MockAnalysis()
    lims_api = MockLimsAPI(samples=lims_samples)

    return UploadScoutAPI(
        hk_api=housekeeper_api,
        scout_api=scout_api,
        madeline_api=madeline_api,
        analysis_api=analysis_mock,
        lims_api=lims_api,
    )


@pytest.fixture(name="upload_mip_analysis_scout_api")
def fixture_upload_mip_analysis_scout_api(
    scout_api: MockScoutAPI,
    madeline_api: MockMadelineAPI,
    lims_samples: List[dict],
    mip_analysis_hk_api: MockHousekeeperAPI,
) -> UploadScoutAPI:
    """Fixture for upload_scout_api"""
    analysis_mock = MockAnalysis()
    lims_api = MockLimsAPI(samples=lims_samples)

    _api = UploadScoutAPI(
        hk_api=mip_analysis_hk_api,
        scout_api=scout_api,
        madeline_api=madeline_api,
        analysis_api=analysis_mock,
        lims_api=lims_api,
    )

    yield _api


@pytest.fixture(name="upload_balsamic_analysis_scout_api")
def fixture_upload_balsamic_analysis_scout_api(
    scout_api: MockScoutAPI,
    madeline_api: MockMadelineAPI,
    lims_samples: List[dict],
    balsamic_analysis_hk_api: MockHousekeeperAPI,
) -> UploadScoutAPI:
    """Fixture for upload_scout_api"""
    analysis_mock = MockAnalysis()
    lims_api = MockLimsAPI(samples=lims_samples)

    _api = UploadScoutAPI(
        hk_api=balsamic_analysis_hk_api,
        scout_api=scout_api,
        madeline_api=madeline_api,
        analysis_api=analysis_mock,
        lims_api=lims_api,
    )

    yield _api



from pathlib import Path

from cg.models.demultiplex.flowcell import Flowcell


def test_flowcell_id(flowcell_path: Path):
    # GIVEN the path to a finished flowcell run
    # GIVEN the flowcell id
    flowcell_id: str = flowcell_path.name.split("_")[-1][1:]

    # WHEN instantiating a flowcell object
    flowcell_obj = Flowcell(flowcell_path)

    # THEN assert that the flowcell flowcell id is correcly parsed
    assert flowcell_obj.flowcell_id == flowcell_id


def test_flowcell_position(flowcell_path: Path):
    # GIVEN the path to a finished flowcell
    # GIVEN a flowcell object
    flowcell_obj = Flowcell(flowcell_path)

    # WHEN fetching the flowcell position
    position = flowcell_obj.flowcell_position

    # THEN assert it is A or B
    assert position in ["A", "B"]


def test_rta_exists(flowcell_object: Flowcell):
    # GIVEN the path to a finished flowcell
    # GIVEN a flowcell object

    # WHEN fetching the path to the RTA file
    rta_file: Path = flowcell_object.rta_complete_path

    # THEN assert that the file exists
    assert rta_file.exists()



"""cg module for cleaning databases and files"""
import logging
from datetime import datetime
from pathlib import Path
from typing import Iterable, List, Optional

import click
from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.apps.scout.scout_export import ScoutExportCase
from cg.apps.scout.scoutapi import ScoutAPI
from cg.cli.workflow.commands import balsamic_past_run_dirs, mip_past_run_dirs, mutant_past_run_dirs
from cg.models.cg_config import CGConfig
from cg.store import Store, models
from housekeeper.store import models as hk_models

LOG = logging.getLogger(__name__)


@click.group()
def clean():
    """Clean up processes"""
    return


clean.add_command(balsamic_past_run_dirs)
clean.add_command(mip_past_run_dirs)
clean.add_command(mutant_past_run_dirs)


@clean.command("hk-alignment-files")
@click.argument("bundle")
@click.option("-y", "--yes", is_flag=True, help="Skip confirmation")
@click.option("-d", "--dry-run", is_flag=True, help="Show files that would be cleaned")
@click.pass_obj
def hk_alignment_files(context: CGConfig, bundle: str, yes: bool = False, dry_run: bool = False):
    """Clean up alignment files in Housekeeper bundle"""
    housekeeper_api: HousekeeperAPI = context.housekeeper_api
    for tag in ["bam", "bai", "bam-index", "cram", "crai", "cram-index"]:

        tag_files = set(housekeeper_api.get_files(bundle=bundle, tags=[tag]))

        if not tag_files:
            LOG.warning(
                "Could not find any files ready for cleaning for bundle %s and tag %s", bundle, tag
            )

        for file_obj in tag_files:
            if not (yes or click.confirm(_get_confirm_question(bundle, file_obj))):
                continue

            file_path: Path = Path(file_obj.full_path)
            if file_obj.is_included and file_path.exists():
                LOG.info("Unlinking %s", file_path)
                if not dry_run:
                    file_path.unlink()

            LOG.info("Deleting %s from database", file_path)
            if not dry_run:
                file_obj.delete()
                housekeeper_api.commit()


def _get_confirm_question(bundle, file_obj):
    if file_obj.is_included:
        question = f"{bundle}: remove file from file system and database: {file_obj.full_path}"
    else:
        question = f"{bundle}: remove file from database: {file_obj.full_path}"
    return question


@clean.command("scout-finished-cases")
@click.option(
    "--days-old",
    type=int,
    default=300,
    help="Clean alignment files with analysis dates older then given number of days",
)
@click.option("-y", "--yes", is_flag=True, help="Skip confirmation")
@click.option("-d", "--dry-run", is_flag=True, help="Shows cases and files that would be cleaned")
@click.pass_context
def scout_finished_cases(
    context: click.Context, days_old: int, yes: bool = False, dry_run: bool = False
):
    """Clean up of solved and archived scout cases"""
    scout_api: ScoutAPI = context.obj.scout_api
    bundles: List[str] = []
    for status in ["archived", "solved"]:
        cases: List[ScoutExportCase] = scout_api.get_cases(status=status, reruns=False)
        cases_added = 0
        for case in cases:
            x_days_ago = datetime.now() - case.analysis_date
            if x_days_ago.days > days_old:
                bundles.append(case.id)
                cases_added += 1
        LOG.info("%s cases marked for bam removal", cases_added)

    for bundle in bundles:
        context.invoke(hk_alignment_files, bundle=bundle, yes=yes, dry_run=dry_run)


@clean.command("hk-past-files")
@click.option("-c", "--case-id", type=str)
@click.option("-t", "--tags", multiple=True)
@click.option("-y", "--yes", is_flag=True, help="skip checks")
@click.option("-d", "--dry-run", is_flag=True, help="Shows cases and files that would be cleaned")
@click.pass_obj
def hk_past_files(context: CGConfig, case_id: str, tags: list, yes: bool, dry_run: bool):
    """Remove files found in older housekeeper bundles"""

    housekeeper_api: HousekeeperAPI = context.housekeeper_api
    status_db: Store = context.status_db

    cases: Iterable[models.Family]
    if case_id:
        cases: List[models.Family] = [status_db.family(case_id)]
    else:
        cases: Iterable[models.Family] = status_db.families()
    for case in cases:
        case_id = case.internal_id
        last_version: Optional[hk_models.Version] = housekeeper_api.last_version(bundle=case_id)
        if not last_version:
            continue
        last_version_file_paths: List[Path] = [
            Path(hk_file.full_path)
            for hk_file in housekeeper_api.get_files(
                bundle=case_id, tags=None, version=last_version.id
            )
        ]
        LOG.info("Searching %s bundle for outdated files", case_id)
        hk_files: List[hk_models.File] = []
        for tag in tags:
            hk_files.extend(housekeeper_api.get_files(bundle=case_id, tags=[tag]))
        for hk_file in hk_files:
            file_path = Path(hk_file.full_path)
            if file_path in last_version_file_paths:
                continue
            LOG.info("Will remove %s", file_path)
            if (yes or click.confirm("Do you want to remove this file?")) and not dry_run:
                hk_file.delete()
                housekeeper_api.commit()
                if file_path.exists():
                    file_path.unlink()
                LOG.info("File removed")



import logging
from typing import Optional

from cg.apps.cgstats.stats import StatsAPI
from cg.apps.coverage import ChanjoAPI
from cg.apps.crunchy import CrunchyAPI
from cg.apps.demultiplex.demultiplex_api import DemultiplexingAPI
from cg.apps.gt import GenotypeAPI
from cg.apps.hermes.hermes_api import HermesApi
from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.apps.lims import LimsAPI
from cg.apps.loqus import LoqusdbAPI
from cg.apps.madeline.api import MadelineAPI
from cg.apps.mutacc_auto import MutaccAutoAPI
from cg.apps.scout.scoutapi import ScoutAPI
from cg.apps.shipping import ShippingAPI
from cg.apps.tb import TrailblazerAPI
from cg.apps.vogue import VogueAPI
from cg.store import Store
from pydantic import BaseModel, EmailStr, Field
from typing_extensions import Literal

LOG = logging.getLogger(__name__)


class Sequencers(BaseModel):
    hiseqx: str
    hiseqga: str
    novaseq: str


class BackupConfig(BaseModel):
    root: Sequencers


class SlurmConfig(BaseModel):
    account: str
    mail_user: EmailStr
    conda_env: Optional[str]
    qos: Literal["low", "normal", "high"] = "low"


class HousekeeperConfig(BaseModel):
    database: str
    root: str


class DemultiplexConfig(BaseModel):
    run_dir: str  # Base path to  un demultiplexed flowcells
    out_dir: str  # Base path to where the demultiplexed results lives
    slurm: SlurmConfig


class TrailblazerConfig(BaseModel):
    service_account: str
    service_account_auth_file: str
    host: str


class CommonAppConfig(BaseModel):
    binary_path: str
    config_path: Optional[str]
    deploy_config: Optional[str]


class FluffyConfig(CommonAppConfig):
    root_dir: str


class LimsConfig(BaseModel):
    host: str
    username: str
    password: str


class CrunchyConfig(BaseModel):
    cram_reference: str
    slurm: SlurmConfig


class MutaccAutoConfig(CommonAppConfig):
    padding: int = 300


class BalsamicConfig(CommonAppConfig):
    root: str
    balsamic_cache: str
    binary_path: str
    conda_env: str
    slurm: SlurmConfig


class MutantConfig(BaseModel):
    root: str
    binary_path: str
    conda_env: str


class MipConfig(BaseModel):
    conda_env: str
    mip_config: str
    pipeline: str
    root: str
    script: str


class CGStatsConfig(BaseModel):
    database: str
    root: str


class MicrosaltConfig(BaseModel):
    root: str
    queries_path: str
    binary_path: str
    conda_env: str


class GisaidConfig(CommonAppConfig):
    submitter: str


class ShippingConfig(CommonAppConfig):
    host_config: str


class DataDeliveryConfig(BaseModel):
    destination_path: str
    covid_destination_path: str
    covid_source_path = str
    covid_report_path: str


class CGConfig(BaseModel):
    database: str
    environment: Literal["production", "stage"] = "stage"
    madeline_exe: str
    bed_path: str
    delivery_path: str
    max_flowcells: Optional[int]
    # Base APIs that always should exist
    status_db_: Store = None
    housekeeper: HousekeeperConfig
    housekeeper_api_: HousekeeperAPI = None

    # App APIs that can be instantiated in CGConfig
    backup: BackupConfig = None
    cgstats: CGStatsConfig = None
    cg_stats_api_: StatsAPI = None
    chanjo: CommonAppConfig = None
    chanjo_api_: ChanjoAPI = None
    crunchy: CrunchyConfig = None
    crunchy_api_: CrunchyAPI = None
    data_delivery: DataDeliveryConfig = Field(None, alias="data-delivery")
    demultiplex: DemultiplexConfig = None
    demultiplex_api_: DemultiplexingAPI = None
    genotype: CommonAppConfig = None
    genotype_api_: GenotypeAPI = None
    hermes: CommonAppConfig = None
    hermes_api_: HermesApi = None
    lims: LimsConfig = None
    lims_api_: LimsAPI = None
    loqusdb: CommonAppConfig = None
    loqusdb_api_: LoqusdbAPI = None
    loqusdb_wes: CommonAppConfig = Field(None, alias="loqusdb-wes")
    madeline_api_: MadelineAPI = None
    mutacc_auto: MutaccAutoConfig = Field(None, alias="mutacc-auto")
    mutacc_auto_api_: MutaccAutoAPI = None
    scout: CommonAppConfig = None
    scout_api_: ScoutAPI = None
    shipping: ShippingConfig = None
    shipping_api_: ShippingAPI = None
    trailblazer: TrailblazerConfig = None
    trailblazer_api_: TrailblazerAPI = None
    vogue: CommonAppConfig = None
    vogue_api_: VogueAPI = None

    # Meta APIs that will use the apps from CGConfig
    balsamic: BalsamicConfig = None
    fluffy: FluffyConfig = None
    microsalt: MicrosaltConfig = None
    gisaid: GisaidConfig = None
    mip_rd_dna: MipConfig = Field(None, alias="mip-rd-dna")
    mip_rd_rna: MipConfig = Field(None, alias="mip-rd-rna")
    mutant: MutantConfig = None

    # These are meta APIs that gets instantiated in the code
    meta_apis: dict = {}

    class Config:
        arbitrary_types_allowed = True
        fields = {
            "cg_stats_api_": "cg_stats_api",
            "chanjo_api_": "chanjo_api",
            "crunchy_api_": "crunchy_api",
            "demultiplex_api_": "demultiplex_api",
            "genotype_api_": "genotype_api",
            "hermes_api_": "hermes_api",
            "housekeeper_api_": "housekeeper_api",
            "lims_api_": "lims_api",
            "loqusdb_api_": "loqusdb_api",
            "madeline_api_": "madeline_api",
            "mutacc_auto_api_": "mutacc_auto_api",
            "scout_api_": "scout_api",
            "shipping_api_": "shipping_api",
            "status_db_": "status_db",
            "trailblazer_api_": "trailblazer_api",
            "vogue_api_": "vogue_api",
        }

    @property
    def status_db(self) -> Store:
        status_db = self.__dict__.get("status_db_")
        if status_db is None:
            LOG.debug("Instantiating status db")
            status_db = Store(self.database)
            self.status_db_ = status_db
        return status_db

    @property
    def housekeeper_api(self) -> HousekeeperAPI:
        housekeeper_api = self.__dict__.get("housekeeper_api_")
        if housekeeper_api is None:
            LOG.debug("Instantiating housekeeper api")
            housekeeper_api = HousekeeperAPI(config=self.dict())
            self.housekeeper_api_ = housekeeper_api
        return housekeeper_api

    @property
    def demultiplex_api(self) -> DemultiplexingAPI:
        demultiplex_api = self.__dict__.get("demultiplex_api_")
        if demultiplex_api is None:
            LOG.debug("Instantiating demultiplexing api")
            demultiplex_api = DemultiplexingAPI(config=self.dict())
            self.demultiplex_api_ = demultiplex_api
        return demultiplex_api

    @property
    def hermes_api(self) -> HermesApi:
        hermes_api = self.__dict__.get("hermes_api_")
        if hermes_api is None:
            LOG.debug("Instantiating hermes api")
            hermes_api = HermesApi(config=self.dict())
            self.hermes_api_ = hermes_api
        return hermes_api

    @property
    def scout_api(self) -> ScoutAPI:
        api = self.__dict__.get("scout_api_")
        if api is None:
            LOG.debug("Instantiating scout api")
            api = ScoutAPI(config=self.dict())
            self.scout_api_ = api
        return api

    @property
    def trailblazer_api(self) -> TrailblazerAPI:
        api = self.__dict__.get("trailblazer_api_")
        if api is None:
            LOG.debug("Instantiating trailblazer api")
            api = TrailblazerAPI(config=self.dict())
            self.trailblazer_api_ = api
        return api

    @property
    def lims_api(self) -> LimsAPI:
        api = self.__dict__.get("lims_api_")
        if api is None:
            LOG.debug("Instantiating lims api")
            api = LimsAPI(config=self.dict())
            self.lims_api_ = api
        return api

    @property
    def vogue_api(self) -> VogueAPI:
        api = self.__dict__.get("vogue_api_")
        if api is None:
            LOG.debug("Instantiating vogue api")
            api = VogueAPI(config=self.dict())
            self.vogue_api_ = api
        return api

    @property
    def crunchy_api(self) -> CrunchyAPI:
        api = self.__dict__.get("crunchy_api_")
        if api is None:
            LOG.debug("Instantiating crunchy api")
            api = CrunchyAPI(config=self.dict())
            self.crunchy_api_ = api
        return api

    @property
    def madeline_api(self) -> MadelineAPI:
        api = self.__dict__.get("madeline_api_")
        if api is None:
            LOG.debug("Instantiating madeline api")
            api = MadelineAPI(config=self.dict())
            self.madeline_api_ = api
        return api

    @property
    def mutacc_auto_api(self) -> MutaccAutoAPI:
        api = self.__dict__.get("mutacc_auto_api_")
        if api is None:
            LOG.debug("Instantiating mutacc_auto api")
            api = MutaccAutoAPI(config=self.dict())
            self.mutacc_auto_api_ = api
        return api

    @property
    def genotype_api(self) -> GenotypeAPI:
        api = self.__dict__.get("genotype_api_")
        if api is None:
            LOG.debug("Instantiating genotype api")
            api = GenotypeAPI(config=self.dict())
            self.genotype_api_ = api
        return api

    @property
    def chanjo_api(self) -> ChanjoAPI:
        api = self.__dict__.get("chanjo_api_")
        if api is None:
            LOG.debug("Instantiating chanjo api")
            api = ChanjoAPI(config=self.dict())
            self.chanjo_api_ = api
        return api

    @property
    def shipping_api(self) -> ShippingAPI:
        api = self.__dict__.get("shipping_api_")
        if api is None:
            LOG.debug("Instantiating shipping api")
            api = ShippingAPI(config=self.shipping.dict())
            self.shipping_api_ = api
        return api

    @property
    def cg_stats_api(self) -> StatsAPI:
        api = self.__dict__.get("cg_stats_api_")
        if api is None:
            LOG.debug("Instantiating cg_stats api")
            api = StatsAPI(config=self.dict())
            self.cg_stats_api_ = api
        return api

    @property
    def loqusdb_api(self) -> LoqusdbAPI:
        api = self.__dict__.get("loqusdb_api_")
        if api is None:
            LOG.debug("Instantiating loqusdb api")
            api = LoqusdbAPI(config=self.dict())
            self.loqusdb_api_ = api
        return api



"""CLI for delivering files with CG"""
import logging
from pathlib import Path
from typing import List, Optional

import click
from cg.constants.delivery import PIPELINE_ANALYSIS_OPTIONS, PIPELINE_ANALYSIS_TAG_MAP
from cg.meta.deliver import DeliverAPI
from cg.meta.rsync import RsyncAPI
from cg.models.cg_config import CGConfig
from cg.store import Store, models

LOG = logging.getLogger(__name__)


@click.group()
def deliver():
    """Deliver files with CG."""
    LOG.info("Running CG deliver")


@deliver.command(name="analysis")
@click.option("-c", "--case-id", help="Deliver the files for a specific case")
@click.option(
    "-t", "--ticket-id", type=int, help="Deliver the files for ALL cases connected to a ticket"
)
@click.option("-d", "--delivery-type", type=click.Choice(PIPELINE_ANALYSIS_OPTIONS), required=True)
@click.option("--dry-run", is_flag=True)
@click.pass_obj
def deliver_analysis(
    context: CGConfig,
    case_id: Optional[str],
    ticket_id: Optional[int],
    delivery_type: str,
    dry_run: bool,
):
    """Deliver analysis files to customer inbox

    Files can be delivered either on case level or for all cases connected to a ticket.
    Any of those needs to be specified.
    """
    if not (case_id or ticket_id):
        LOG.info("Please provide a case-id or ticket-id")
        return

    inbox: str = context.delivery_path
    if not inbox:
        LOG.info("Please specify the root path for where files should be delivered")
        return

    status_db: Store = context.status_db
    deliver_api = DeliverAPI(
        store=status_db,
        hk_api=context.housekeeper_api,
        case_tags=PIPELINE_ANALYSIS_TAG_MAP[delivery_type]["case_tags"],
        sample_tags=PIPELINE_ANALYSIS_TAG_MAP[delivery_type]["sample_tags"],
        project_base_path=Path(inbox),
        delivery_type=delivery_type,
    )
    deliver_api.set_dry_run(dry_run)
    cases: List[models.Family] = []
    if case_id:
        case_obj: models.Family = status_db.family(case_id)
        if not case_obj:
            LOG.warning("Could not find case %s", case_id)
            return
        cases.append(case_obj)
    else:
        cases = status_db.get_cases_from_ticket(ticket_id=ticket_id).all()
        if not cases:
            LOG.warning("Could not find cases for ticket_id %s", ticket_id)
            return

    for case_obj in cases:
        deliver_api.deliver_files(case_obj=case_obj)


@deliver.command(name="rsync")
@click.argument("ticket_id", type=int, required=True)
@click.option("--dry-run", is_flag=True)
@click.pass_obj
def rsync(context: CGConfig, ticket_id: int, dry_run: bool):
    """The folder generated using the "cg deliver analysis" command will be
    rsynced with this function to the customers inbox on caesar.
    """
    rsync_api = RsyncAPI(config=context)
    rsync_api.run_rsync_command(ticket_id=ticket_id, dry_run=dry_run)



"""Handler to find business data objects"""
import datetime as dt
from typing import List, Optional

from sqlalchemy import and_, func, or_
from sqlalchemy.orm import Query

from cg.constants import Pipeline
from cg.store import models
from cg.store.api.base import BaseHandler


class FindBusinessDataHandler(BaseHandler):
    """Contains methods to find business data model instances"""

    def analyses(self, *, family: models.Family = None, before: dt.datetime = None) -> Query:
        """Fetch multiple analyses."""
        records = self.Analysis.query
        if family:
            query_family = family
            records = records.filter(models.Analysis.family == query_family)
        if before:
            subq = (
                self.Analysis.query.join(models.Analysis.family)
                .filter(models.Analysis.started_at < before)
                .group_by(models.Family.id)
                .with_entities(
                    models.Analysis.family_id,
                    func.max(models.Analysis.started_at).label("started_at"),
                )
                .subquery()
            )
            records = records.join(
                subq,
                and_(
                    self.Analysis.family_id == subq.c.family_id,
                    self.Analysis.started_at == subq.c.started_at,
                ),
            ).filter(models.Analysis.started_at < before)
        return records

    def analyses_ready_for_vogue_upload(
        self,
        completed_after: Optional[dt.date],
        completed_before: Optional[dt.date],
    ) -> Query:
        """Fetch all cases with a finished analysis that has not been uploaded to Vogue.
        Optionally fetch those cases finished before and/or after a specified date"""
        # records = self.Family.query.join(models.Family.analyses).filter(
        #     models.Analysis.uploaded_to_vogue_at.is_(None)
        # )
        records = self.latest_analyses().filter(models.Analysis.uploaded_to_vogue_at.is_(None))

        if completed_after:
            records = records.filter(models.Analysis.completed_at > completed_after)
        if completed_before:
            records = records.filter(models.Analysis.completed_at < completed_before)

        return records

    def latest_analyses(self) -> Query:
        """Fetch latest analysis for all cases."""

        records = self.Analysis.query
        sub_query = (
            self.Analysis.query.join(models.Analysis.family)
            .group_by(models.Family.id)
            .with_entities(
                models.Analysis.family_id, func.max(models.Analysis.started_at).label("started_at")
            )
            .subquery()
        )
        records = records.join(
            sub_query,
            and_(
                self.Analysis.family_id == sub_query.c.family_id,
                self.Analysis.started_at == sub_query.c.started_at,
            ),
        )
        return records

    def analysis(self, family: models.Family, started_at: dt.datetime) -> models.Analysis:
        """Fetch an analysis."""
        return self.Analysis.query.filter_by(family=family, started_at=started_at).first()

    def deliveries(self) -> Query:
        """Fetch all deliveries."""
        return self.Delivery.query

    def families(
        self, *, customers: [models.Customer] = None, enquiry: str = None, action: str = None
    ) -> Query:
        """Fetch families."""

        records = self.Family.query

        if customers:
            customer_ids = []
            for customer in customers:
                customer_ids.append(customer.id)
            records = records.filter(models.Family.customer_id.in_(customer_ids))

        records = (
            records.filter(
                or_(
                    models.Family.name.like(f"%{enquiry}%"),
                    models.Family.internal_id.like(f"%{enquiry}%"),
                )
            )
            if enquiry
            else records
        )

        records = records.filter_by(action=action) if action else records
        return records.order_by(models.Family.created_at.desc())

    def families_in_customer_group(
        self, *, customers: List[models.Customer] = None, enquiry: str = None
    ) -> Query:
        """Fetch all families including those from collaborating customers."""
        records = self.Family.query.join(models.Family.customer, models.Customer.customer_group)

        if customers:
            customer_group_ids = []
            for customer in customers:
                customer_group_ids.append(customer.customer_group_id)
            records = records.filter(models.CustomerGroup.id.in_(customer_group_ids))

        records = (
            records.filter(
                or_(
                    models.Family.name.like(f"%{enquiry}%"),
                    models.Family.internal_id.like(f"%{enquiry}%"),
                )
            )
            if enquiry
            else records
        )

        return records.order_by(models.Family.created_at.desc())

    def family(self, internal_id: str) -> models.Family:
        """Fetch a family by internal id from the database."""
        return self.Family.query.filter_by(internal_id=internal_id).first()

    def family_samples(self, family_id: str) -> List[models.FamilySample]:
        """Find the samples of a family."""
        return (
            self.FamilySample.query.join(models.FamilySample.family, models.FamilySample.sample)
            .filter(models.Family.internal_id == family_id)
            .all()
        )

    def get_samples_by_family_id(self, family_id: str) -> List[models.Sample]:
        """Get samples on a given family_id"""
        case: models.Family = self.family(internal_id=family_id)
        return [link.sample for link in case.links] if case else []

    def get_sequenced_samples(self, family_id: str) -> List[models.Sample]:
        """Get sequenced samples by family_id"""

        samples: List[models.Sample] = self.get_samples_by_family_id(family_id)
        return [sample for sample in samples if sample.sequencing_qc]

    def find_family(self, customer: models.Customer, name: str) -> models.Family:
        """Find a family by family name within a customer."""
        return self.Family.query.filter_by(customer=customer, name=name).first()

    def find_family_by_name(self, name: str) -> models.Family:
        """Find a family by family name within a customer."""
        return self.Family.query.filter_by(name=name).first()

    def find_sample(self, customer: models.Customer, name: str) -> Query:
        """Find samples within a customer."""
        return self.Sample.query.filter_by(customer=customer, name=name)

    def flowcell(self, name: str) -> models.Flowcell:
        """Fetch a flowcell."""
        return self.Flowcell.query.filter(models.Flowcell.name == name).first()

    def flowcells(
        self, *, status: str = None, family: models.Family = None, enquiry: str = None
    ) -> Query:
        """Fetch all flowcells."""
        records = self.Flowcell.query
        if family:
            records = records.join(models.Flowcell.samples, models.Sample.links).filter(
                models.FamilySample.family == family
            )
        if status:
            records = records.filter_by(status=status)
        if enquiry:
            records = records.filter(models.Flowcell.name.like(f"%{enquiry}%"))
        return records.order_by(models.Flowcell.sequenced_at.desc())

    def invoices(self, invoiced: bool = None) -> Query:
        """Fetch invoices."""
        query = self.Invoice.query
        if invoiced:
            query = query.filter(models.Invoice.invoiced_at.isnot(None))
        else:
            query = query.filter(models.Invoice.invoiced_at.is_(None))
        return query

    def invoice(self, invoice_id: int) -> models.Invoice:
        """Fetch an invoice."""
        return self.Invoice.get(invoice_id)

    def invoice_samples(self, *, invoice_id: int = None) -> Query:
        """Fetch pools and samples for an invoice"""
        pools = self.Pool.query.filter_by(invoice_id=invoice_id).all()
        samples = self.Sample.query.filter_by(invoice_id=invoice_id).all()
        return pools + samples

    def link(self, family_id: str, sample_id: str) -> models.FamilySample:
        """Find a link between a family and a sample."""
        return (
            self.FamilySample.query.join(models.FamilySample.family, models.FamilySample.sample)
            .filter(models.Family.internal_id == family_id, models.Sample.internal_id == sample_id)
            .first()
        )

    def links(self, case_id: str, sample_id: str, ticket: int) -> Query:
        """Find a link between a family and a sample."""

        query = self.FamilySample.query.join(models.FamilySample.family, models.FamilySample.sample)

        if case_id:
            query = query.filter(models.Family.internal_id == case_id)

        if sample_id:
            query = query.filter(models.Sample.internal_id == sample_id)

        if ticket:
            query = query.filter(models.Sample.ticket_number == ticket)

        return query

    def new_invoice_id(self) -> int:
        """Fetch invoices."""
        query = self.Invoice.query.all()
        ids = [inv.id for inv in query]
        return max(ids) + 1 if ids else 0

    def pools(
        self, *, customers: Optional[List[models.Customer]] = None, enquiry: str = None
    ) -> Query:
        """Fetch all the pools for a customer."""
        records = self.Pool.query

        if customers:
            customer_ids = []
            for customer in customers:
                customer_ids.append(customer.id)
            records = records.filter(models.Pool.customer_id.in_(customer_ids))

        records = (
            records.filter(
                or_(models.Pool.name.like(f"%{enquiry}%"), models.Pool.order.like(f"%{enquiry}%"))
            )
            if enquiry
            else records
        )

        return records.order_by(models.Pool.created_at.desc())

    def pool(self, pool_id: int) -> models.Pool:
        """Fetch a pool."""
        return self.Pool.get(pool_id)

    def sample(self, internal_id: str) -> models.Sample:
        """Fetch a sample by lims id."""
        return self.Sample.query.filter_by(internal_id=internal_id).first()

    def samples(
        self, *, customers: Optional[List[models.Customer]] = None, enquiry: str = None
    ) -> Query:
        records = self.Sample.query

        if customers:
            customer_ids = []
            for customer in customers:
                customer_ids.append(customer.id)
            records = records.filter(models.Sample.customer_id.in_(customer_ids))

        records = (
            records.filter(
                or_(
                    models.Sample.name.like(f"%{enquiry}%"),
                    models.Sample.internal_id.like(f"%{enquiry}%"),
                )
            )
            if enquiry
            else records
        )
        return records.order_by(models.Sample.created_at.desc())

    def samples_by_ids(self, **identifiers) -> Query:
        records = self.Sample.query

        for identifier_name, identifier_value in identifiers.items():
            identifier = getattr(models.Sample, identifier_name)
            records = records.filter(identifier.contains(identifier_value))

        return records.order_by(models.Sample.internal_id.asc())

    def samples_in_customer_group(
        self, *, customers: Optional[List[models.Customer]] = None, enquiry: str = None
    ) -> Query:
        """Fetch all samples including those from collaborating customers."""

        records = self.Sample.query.join(models.Sample.customer, models.Customer.customer_group)

        if customers:
            customer_group_ids = []
            for customer in customers:
                customer_group_ids.append(customer.customer_group_id)
            records = records.filter(models.CustomerGroup.id.in_(customer_group_ids))

        records = (
            records.filter(
                or_(
                    models.Sample.name.like(f"%{enquiry}%"),
                    models.Sample.internal_id.like(f"%{enquiry}%"),
                )
            )
            if enquiry
            else records
        )
        return records.order_by(models.Sample.created_at.desc())



"""Tests for the demultiplex flowcell class"""
from pathlib import Path

import pytest
from cg.models.demultiplex.flowcell import Flowcell


def test_get_run_parameters_when_non_existing(fixtures_dir: Path):
    # GIVEN a flowcell object with a directory without run parameters
    flowcell = Flowcell(flowcell_path=fixtures_dir)
    assert flowcell.run_parameters_path.exists() is False

    # WHEN fetching the run parameters object
    with pytest.raises(FileNotFoundError):
        # THEN assert that a FileNotFound error is raised
        flowcell.run_parameters_object



import copy
import datetime as dt
from pathlib import Path

import pytest
import yaml
from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.constants import Pipeline
from cg.meta.workflow.mip_dna import MipDNAAnalysisAPI
from cg.meta.workflow.mip_rna import MipRNAAnalysisAPI
from cg.models.cg_config import CGConfig
from tests.store_helpers import StoreHelpers


@pytest.fixture(name="mip_case_ids")
def fixture_mip_case_ids() -> dict:
    """Dictionary of case ids, connected samples, their name and if they should fail (textbook or not)"""

    return {
        "yellowhog": {
            "textbook": True,
            "name": "F0000001",
            "internal_id": "ACC00000",
        },
        "purplesnail": {
            "textbook": False,
            "name": "F0000003",
            "internal_id": "ACC00000",
        },
        "bluezebra": {
            "textbook": True,
            "name": "F0000002",
            "internal_id": "ACC00000",
        },
    }


@pytest.fixture(name="mip_case_dirs")
def fixture_mip_case_dirs(mip_case_ids: dict, project_dir: Path) -> dict:
    """Create case directories and return a dictionary of the tmpdir paths"""

    mip_case_dirs = {}

    for case in mip_case_ids:
        case_dir = Path(project_dir / "cases" / case)
        case_dir.mkdir(exist_ok=True, parents=True)
        mip_case_dirs[case] = case_dir

    return mip_case_dirs


@pytest.fixture(name="mip_deliverables")
def fixture_mip_deliverables(
    mip_case_ids: dict,
    mip_case_dirs: dict,
    mip_deliverables_file: Path,
) -> dict:
    """Create deliverables for mip store testing"""

    # Create a dict of paths to the deliverables to be used in later stages
    deliverables_paths = {}

    # Load general deliverables file
    with open(mip_deliverables_file, "r") as mip_dna_deliverables:
        mip_deliverables = yaml.full_load(mip_dna_deliverables)

    for case, value in mip_case_ids.items():
        if value["textbook"]:
            case_specific_deliverables = copy.deepcopy(mip_deliverables)

            for file in case_specific_deliverables["files"]:
                if file["path_index"]:
                    file["path_index"] = str(mip_case_dirs[case] / Path(file["path_index"]).name)
                    # Touch the tmp file
                    open(file["path_index"], "a").close()
                file["path"] = str(mip_case_dirs[case] / Path(file["path"]).name)
                # Touch the tmp file
                open(file["path"], "a").close()

            case_deliverables_path = mip_case_dirs[case] / f"{case}_deliverables.yaml"
            with open(case_deliverables_path, "w") as fh:
                yaml.dump(case_specific_deliverables, fh)
        else:
            case_deliverables_path = mip_case_dirs[case] / f"{case}_deliverables.yaml"
            with open(case_deliverables_path, "w") as fh:
                yaml.dump({"files": []}, fh)
        deliverables_paths[case] = case_deliverables_path

    return deliverables_paths


@pytest.fixture(name="mip_qc_sample_info")
def fixture_mip_qc_sample_info(mip_case_ids: dict, mip_case_dirs: dict) -> dict:
    """Create qc sample info files for cases"""
    # Store paths in a list to be used in other stages

    qc_sample_info = {}

    for case in mip_case_ids:
        sample_info_data = {
            "date": dt.datetime.now(),
            "analysisrunstatus": "finished",
            "analysis_date": dt.datetime.now(),
            "case": case,
            "mip_version": "latest",
        }

        sample_info_path = mip_case_dirs[case] / f"{case}_qc_sample_info.yaml"
        qc_sample_info[case] = sample_info_path
        with open(sample_info_path, "w") as fh:
            yaml.dump(sample_info_data, fh)

    return qc_sample_info


@pytest.fixture(name="mip_configs")
def fixture_mip_configs(
    mip_case_ids: dict, mip_case_dirs: dict, mip_deliverables: dict, mip_qc_sample_info: dict
) -> dict:
    """Create config files for mip case"""
    # Storing paths in a list to be returned and used in later stages.

    config_dict = {}

    for case in mip_case_ids:
        config = {
            "email": "fake_email@notscilifelab.com",
            "case_id": case,
            "samples": "dummy_samples",
            "is_dryrun": False,
            "outdata_dir": str(mip_case_dirs[case]),
            "priority": "medium",
            "sampleinfo_path": str(mip_qc_sample_info[case]),
            "sample_info_file": str(mip_qc_sample_info[case]),
            "analysis_type": {mip_case_ids[case]["internal_id"]: "wgs"},
            "slurm_quality_of_service": "medium",
            "store_file": str(mip_deliverables[case]),
        }
        analysis_path = Path(mip_case_dirs[case] / "analysis")
        Path(analysis_path).mkdir(parents=True, exist_ok=True)
        config_path = analysis_path / f"{case}_config.yaml"
        config_dict[case] = config_path
        with open(config_path, "w") as fh:
            yaml.dump(config, fh)

    return config_dict


@pytest.fixture(name="rna_mip_context")
def fixture_rna_mip_context(
    cg_context: CGConfig,
    analysis_family_single_case: dict,
    helpers: StoreHelpers,
    apptag_rna: str,
    case_id: str,
    housekeeper_api: HousekeeperAPI,
) -> CGConfig:
    cg_context.housekeeper_api_ = housekeeper_api
    analysis_family_single_case["data_analysis"] = str(Pipeline.MIP_RNA)
    if not cg_context.status_db.family(case_id):
        helpers.ensure_case_from_dict(
            cg_context.status_db, case_info=analysis_family_single_case, app_tag=apptag_rna
        )
    cg_context.meta_apis["analysis_api"] = MipRNAAnalysisAPI(cg_context)
    return cg_context


@pytest.fixture(name="dna_mip_context")
def fixture_dna_mip_context(
    cg_context: CGConfig, helpers: StoreHelpers, mip_case_ids: dict, housekeeper_api: HousekeeperAPI
) -> CGConfig:
    _store = cg_context.status_db
    cg_context.housekeeper_api_ = housekeeper_api

    # Add apptag to db
    helpers.ensure_application_version(store=_store, application_tag="WGSA", application_type="wgs")

    # Add sample, cases and relationships to db

    for case_id in mip_case_ids:
        if not _store.family(case_id):
            case_obj = helpers.add_case(
                store=_store,
                internal_id=case_id,
                case_id=mip_case_ids[case_id]["name"],
            )
            sample = helpers.add_sample(
                store=_store,
                sample=mip_case_ids[case_id]["internal_id"],
                data_analysis=Pipeline.MIP_DNA,
                customer_name="cust000",
                application_tag="WGSA",
                application_type="wgs",
                gender="unknown",
            )
            helpers.add_relationship(store=_store, sample=sample, case=case_obj, status="affected")
    cg_context.meta_apis["analysis_api"] = MipDNAAnalysisAPI(cg_context)
    return cg_context



from pathlib import Path
from typing import Optional
from pydantic import BaseModel, validator

from cg.meta.upload.gisaid.constants import AUTHORS


class FastaFile(BaseModel):
    header: str
    sequence: str


class UpploadFiles(BaseModel):
    csv_file: Path
    fasta_file: Path


class Lab(BaseModel):
    city: str
    address: str
    region_nr: str
    institute: str


class GisaidSample(BaseModel):
    family_id: str
    cg_lims_id: str
    lab: Lab
    submitter: str
    fn: str
    covv_collection_date: str
    covv_subm_sample_id: str
    covv_virus_name: Optional[str]
    covv_orig_lab: Optional[str]
    covv_type: Optional[str] = "betacoronavirus"
    covv_passage: Optional[str] = "Original"
    covv_location: Optional[str]
    covv_host: Optional[str] = "Human"
    covv_gender: Optional[str] = "unknown"
    covv_patient_age: Optional[str] = "unknown"
    covv_patient_status: Optional[str] = "unknown"
    covv_seq_technology: Optional[str] = "Illumina NovaSeq"
    covv_orig_lab_addr: Optional[str]
    covv_subm_lab: Optional[str] = "Karolinska University Hospital"
    covv_subm_lab_addr: Optional[str] = "171 76 Stockholm, Sweden"
    covv_authors: Optional[str] = " ,".join(AUTHORS)

    @validator("lab", pre=True)
    def lab_info(cls, v):
        if v == "Stockholm":
            return Lab(
                city="Stockholm",
                address="171 76 Stockholm, Sweden",
                region_nr="01",
                institute="Karolinska University Hospital",
            )
        elif v == "Visby":
            return Lab(
                city="Visby",
                address="621 84 Visby, Sweden",
                region_nr="09",
                institute="LaboratorieMedicinskt Centrum Gotland",
            )
        raise ValueError("must be Stockholm or Visby")

    @validator("covv_location", always=True)
    def parse_location(cls, v, values):
        lab = values.get("lab")
        return f"Europe/Sweden/{lab.city}"

    @validator("covv_orig_lab_addr", always=True)
    def parse_orig_lab_addr(cls, v, values):
        lab = values.get("lab")
        return lab.address

    @validator("covv_orig_lab", always=True)
    def parse_orig_lab(cls, v, values):
        lab = values.get("lab")
        return lab.institute

    @validator("covv_subm_sample_id", always=True)
    def parse_subm_sample_id(cls, v, values):
        lab = values.get("lab")
        return f"{lab.region_nr}_SE100_{v}"

    @validator("covv_virus_name", always=True)
    def parse_virus_name(cls, v, values):
        sample_name = values.get("covv_subm_sample_id")
        date = values.get("covv_collection_date")
        return f"hCoV-19/Sweden/{sample_name}/{date}"



from pathlib import Path

import pytest
from cg.models.demultiplex.demux_results import DemuxResults
from cg.models.demultiplex.flowcell import Flowcell
from tests.apps.demultiplex.conftest import fixture_demultiplex_fixtures


@pytest.fixture(name="flowcell_full_name")
def fixture_flowcell_full_name() -> str:
    return "201203_A00689_0200_AHVKJCDRXX"


@pytest.fixture(name="flowcell_runs")
def fixture_flowcell_runs(demultiplex_fixtures: Path) -> Path:
    return demultiplex_fixtures / "flowcell_runs"


@pytest.fixture(name="flowcell_path")
def fixture_flowcell_path(flowcell_runs: Path, flowcell_full_name: str) -> Path:
    return flowcell_runs / flowcell_full_name


@pytest.fixture(name="flowcell_object")
def fixture_flowcell_object(flowcell_path: Path) -> Flowcell:
    return Flowcell(flowcell_path)


@pytest.fixture(name="demultiplexed_runs")
def fixture_demultiplexed_runs(demultiplex_fixtures: Path) -> Path:
    return demultiplex_fixtures / "demultiplexed-runs"


@pytest.fixture(name="demultiplexed_flowcell")
def fixture_demultiplexed_flowcell(demultiplexed_runs: Path, flowcell_full_name: str) -> Path:
    return demultiplexed_runs / flowcell_full_name


@pytest.fixture(name="demux_results")
def fixture_demux_results(demultiplexed_flowcell: Path, flowcell_object: Flowcell) -> DemuxResults:
    return DemuxResults(demux_dir=demultiplexed_flowcell, flowcell=flowcell_object)



"""
    Tests for GisaidAPI
"""
from pathlib import Path
from typing import List

import pytest
from cg.apps.housekeeper.hk import HousekeeperAPI
from cg.exc import FastaSequenceMissingError, HousekeeperVersionMissingError
from cg.meta.upload.gisaid.models import FastaFile, GisaidSample
from tests.store_helpers import StoreHelpers


def test_get_fasta_sequence(valid_fasta_file, gisaid_api):
    # GIVEN a valid consensus fasta file

    # WHEN running get_fasta_sequence
    sequence = gisaid_api.get_fasta_sequence(fastq_path=valid_fasta_file)

    # THEN assert the returned sequence is the second line in the file
    with open(valid_fasta_file) as handle:
        fasta_lines = handle.readlines()
        assert sequence == fasta_lines[1].rstrip("\n")


def test_get_fasta_sequence_invalid_file(gisaid_api, invalid_fasta_file):
    # GIVEN a invalid consensus fasta file
    # WHEN running get_fasta_sequence
    # THEN FastaSequenceMissingError is raised
    with pytest.raises(FastaSequenceMissingError):
        gisaid_api.get_fasta_sequence(fastq_path=invalid_fasta_file)


def test_get_gisaid_fasta_objects(
    gisaid_api,
    four_gisaid_samples: List[GisaidSample],
    bundle_with_four_samples: dict,
    helpers: StoreHelpers,
    real_housekeeper_api: HousekeeperAPI,
):
    # GIVEN four gisaid samples and a hk api populated with a bundle with consensus files related to the samples
    gisaid_api.housekeeper_api_ = helpers.ensure_hk_bundle(
        real_housekeeper_api, bundle_with_four_samples, include=True
    )

    # WHEN running get_gisaid_fasta_objects with the four gisaid samples
    fasta_objects: List[FastaFile] = gisaid_api.get_gisaid_fasta_objects(
        gsaid_samples=four_gisaid_samples
    )
    # THEN files were found and parsed into four fasta file objects
    assert len(fasta_objects) == 4
    for obj in fasta_objects:
        assert isinstance(obj, FastaFile)


def test_get_gisaid_fasta_objects_no_fasta(
    gisaid_api,
    real_housekeeper_api: HousekeeperAPI,
    dummy_gisaid_sample,
):
    # GIVEN a hk api and a dummy gisaid sample with family_id, not represented in the hk api
    gisaid_api.housekeeper_api_ = real_housekeeper_api

    # WHEN running get_gisaid_fasta_objects
    # THEN HousekeeperVersionMissingError is being raised
    with pytest.raises(HousekeeperVersionMissingError):
        gisaid_api.get_gisaid_fasta_objects(gsaid_samples=[dummy_gisaid_sample])


def test_build_gisaid_fasta(
    gisaid_api,
    four_gisaid_samples: List[GisaidSample],
    bundle_with_four_samples: dict,
    helpers: StoreHelpers,
    real_housekeeper_api: HousekeeperAPI,
    valid_concat_fasta_file,
    gisaid_file_name: Path,
):
    # GIVEN four gisaid samples and a hk api populated with a bundle with consensus files related to the samples
    gisaid_api.housekeeper_api_ = helpers.ensure_hk_bundle(
        real_housekeeper_api, bundle_with_four_samples, include=True
    )

    # WHEN running build_gisaid_fasta with the four gisaid samples
    gisaid_api.build_gisaid_fasta(
        gsaid_samples=four_gisaid_samples, file_name=gisaid_file_name.name
    )
    # THEN the concatenated fasta file has the expected content
    assert gisaid_file_name.read_text() == valid_concat_fasta_file


def test_get_gisaid_samples(gisaid_api, case_id):
    # GIVEN a gisaid_api with a Store populated with sequenced samples with family_id=case_id

    # WHEN running get_gisaid_samples
    gisaid_samples = gisaid_api.get_gsaid_samples(family_id=case_id)

    # THEN assert
    assert gisaid_samples != []
    for sample in gisaid_samples:
        assert isinstance(sample, GisaidSample)


def test_get_gisaid_samples_no_samples(gisaid_api):
    # GIVEN a

    # WHEN running get_gisaid_samples
    gisaid_samples = gisaid_api.get_gsaid_samples(family_id="dummy_case")

    # THEN assert
    assert gisaid_samples == []


def test_build_gisaid_csv(
    gisaid_api, four_gisaid_samples, four_samples_csv, gisaid_file_name: Path
):
    # GIVEN four gisaid samples

    # WHEN running build_gisaid_csv
    gisaid_api.build_gisaid_csv(gsaid_samples=four_gisaid_samples, file_name=gisaid_file_name)

    # THEN the generated file should have the expected content
    assert gisaid_file_name.read_text() == four_samples_csv



import logging
from typing import List

from cg.apps.lims import LimsAPI
from cg.models.lims.sample import LimsSample, Udf

LOG = logging.getLogger(__name__)


def build_lims_sample(customer: str, samples: List[dict]) -> List[LimsSample]:
    """Convert order input to lims interface input."""
    samples_lims = []
    for sample in samples:
        LOG.debug(f"{sample['name']}: prepare LIMS input")
        sample["customer"] = customer
        lims_sample: LimsSample = LimsSample.parse_obj(sample)
        udf: Udf = Udf.parse_obj(sample)
        lims_sample.udfs = udf
        samples_lims.append(lims_sample)
    return samples_lims


def process_lims(lims_api: LimsAPI, lims_order: dict, new_samples: List[dict]):
    """Process samples to add them to LIMS."""
    samples_lims: List[LimsSample] = build_lims_sample(lims_order["customer"], samples=new_samples)
    project_name = lims_order.get("ticket", lims_order["name"])
    # Create new lims project
    project_data = lims_api.submit_project(
        project_name, [lims_sample.dict() for lims_sample in samples_lims]
    )
    lims_map = lims_api.get_samples(projectlimsid=project_data["id"], map_ids=True)
    return project_data, lims_map
